# vim: filetype=python sw=4 sts=4 et ai ff=unix fileencoding=utf8 :
import pytest
from inform import os_error
from textwrap import dedent
import subprocess
import os

# set various environment variables so avendesora uses local gpg key and config
# directory rather than the users.
os.environ['HOME'] = 'home'

# change the current working directory to the test directory
cwd = os.getcwd()
if not cwd.endswith('/tests'):
    os.chdir('tests')

# Run avendesora
# Cannot determine whether coverage analysis is needed, so simply always do it
# Whoops, not so fast. The python coverage analysis is broken when it comes to
# the exit status, it always returns 1. So instead, never do it.
def run(args):
    #return subprocess.check_output(['coverage', 'run', '-m'] + args.split())
    return subprocess.check_output(args.split())

# test_add() {{{1
def test_add():
    try:
        result = run('avendesora help add')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Add a new account.

        Usage:
            avendesora add [options] [<template>]
            avendesora a   [options] [<template>]

        Options:
            -f <file>, --file <file>
                                    Add account to specified accounts file.

        Creates a new account starting from a template. The template consists of
        boilerplate code and fields. The fields take the from _NAME_. They
        should be replaced by appropriate values or deleted if not needed. If
        you are using the Vim editor, it is preconfigured to jump to the next
        field when you press 'n'.  If the field is surrounded by '<<' and '>>',
        as in '<<_ACCOUNT_NUMBER_>>', the value you enter will be concealed.

        You can create your own templates by adding them to 'account_templates'
        in the ~/.config/avendesora/config file.

        You can change the editor used when adding account by changing the
        'edit_template', also found in the ~/.config/avendesora/config file.

        The default template is bank. The available templates are:
            bank
            shell
            website

        The available accounts files are (the default is given first):
            accounts
            stealth_accounts
    """).strip()
    assert result.decode('utf-8') == expected

# test_archive() {{{1
def test_archive():
    try:
        result = run('avendesora help archive')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Generates archive of all account information.

        Usage:
            avendesora archive
            avendesora A

        This command creates an encrypted archive that contains all the
        information in your accounts files, including the fully generated
        secrets.  You should never need this file, but its presence protects
        you in case you lose access to Avendesora. To access your secrets
        without Avendesora, simply decrypt the archive file with GPG.  The
        actual secrets will be hidden, but it easy to retrieve them even
        without Avendesora. When hidden, the secrets are encoded in base64.
        You can decode it by running 'base64 -d -' and pasting the encoded
        secret into the terminal.

        When you run this command it overwrites the existing archive. If you
        have accidentally deleted an account or changed a secret, then
        replacing the archive could cause the last copy of the original
        information to be lost. To prevent this from occurring it is a good
        practice to run the 'changed' command before regenerating the
        archive.  It describes all of the changes that have occurred since
        the last time the archive was generated. You should only regenerate
        the archive once you have convinced yourself all of the changes are
        as expected.
    """).strip()
    assert result.decode('utf-8') == expected

# test_browse() {{{1
def test_browse():
    try:
        result = run('avendesora help browse')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Open account URL in web browser.

        Usage:
            avendesora browse [options] <account> [<key>]
            avendesora b      [options] <account> [<key>]

        Options:
            -b <browser>, --browser <browser>
                                    Open account in specified browser.
            -l, --list              List available URLs rather than open first.

        The account is examined for URLs, a URL is chosen, and then that URL
        is opened in the chosen browser.  First URLs are gathered from the
        'urls' account attribute, which can be a string containing one or
        more URLs, a list, or a dictionary.  If 'urls' is a dictionary, the
        desired URL can be chosen by entering the key as an argument to the
        browse command. If a key is not given, then the 'default_url'
        account attribute is used to specify the key to use by default. If
        'urls' is not a dictionary, then the first URL specified is used.
        URLs are also taken from RecognizeURL objects in the 'discovery'
        account attribute.  If the 'name' argument is specified, the
        corresponding URL can be chosen using a key.

        The default browser is x. You can override the default
        browser on a per-account basis by adding an attribute named
        'browser' to the account.  An example of when you would specify the
        browser in an account would be an account associated with Tor hidden
        service, which generally can only be accessed using torbrowser:

            class SilkRoad(Account):
                passcode = Passphrase()
                username = Passphrase(length=2, sep='-')
                url = 'http://silkroad6ownowfk.onion'
                browser = 't'

        The available browsers are:
            c      google-chrome {url}
            ci     google-chrome --incognito {url}
            f      firefox -new-tab {url}
            fp     firefox -private-window {url}
            q      qutebrowser {url}
            t      torbrowser {url}
            x      xdg-open {url}

    """).strip()
    assert result.decode('utf-8') == expected

# test_changed() {{{1
def test_changed():
    try:
        result = run('avendesora help changed')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Show changes since archive was created.

        Usage:
            avendesora changed
            avendesora C

        When you run the 'archive' command it overwrites the existing
        archive. If you have accidentally deleted an account or changed a
        secret, then replacing the archive could cause the last copy of the
        original information to be lost. To prevent this from occurring it
        is a good practice to run the 'changed' command before regenerating
        the archive.  It describes all of the changes that have occurred
        since the last time the archive was generated. You should only
        regenerate the archive once you have convinced yourself all of the
        changes are as expected.
    """).strip()
    assert result.decode('utf-8') == expected

# test_collaborate() {{{1
def test_collaborate():
    try:
        result = run('avendesora help collaborate')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        If you share an accounts file with a partner, then either partner
        can create new secrets and the other partner can reproduce their
        values once a small amount of relatively non-confidential
        information is shared. This works because the security of the
        generated secrets is based on the master seed, and that seed is
        contained in the accounts file that is shared in a secure manner
        once at the beginning.  For example, imagine one partner creates an
        account at the US Postal Service website and then informs the
        partner that the name of the new account is usps and the username is
        justus.  That is enough information for the second partner to
        generate the password and login. And notice that the necessary
        information can be shared over an insecure channel. For example, it
        could be sent in a text message or from a phone where trustworthy
        encryption is not available.

        The first step in using Avendesora to collaborate with a partner is
        for one of the partners to generate and then share an accounts file
        that is dedicated to the shared accounts.  This file contains the
        master seed, and it is critical to keep this value secure. Thus, it
        is recommended that the shared file be encrypted.

        Consider an example where you, Alice, are sharing accounts with your
        business partner, Bob.  You have hired a contractor to run your
        email server, Eve, who unbeknownst to you is reading your email in
        order to steal valuable secrets.  Together, you and Bob jointly run
        Teneya Enterprises. Since you expect more people will need access to
        the accounts in the future, you choose to the name the file after
        the company rather than your partner.  To share accounts with Bob,
        you start by getting Bob's public GPG key.  Then, create the new
        accounts file with something like:

            avendesora new -g alice@teneya.com -g bob@teneya.com teneya.gpg

        This generates a new accounts file, ~/.config/avendesora/teneya.gpg,
        and encrypts it so only you and Bob can open it.  Mail this file to
        Bob. Since it is encrypted, it is to safe to send the file through
        email.  Even though Eve can read this message, the accounts file is
        encrypted so Eve cannot access the master seed it contains.  Bob
        should put the file in ~/.config/avendesora and then add it to
        accounts_files in ~/.config/avendesora/accounts_files.  You are now
        ready to share accounts.

        Then, one partner creates a new account and mails the account entry
        to the other partner.  This entry does not contain enough
        information to allow an eavesdropper such as Eve to be able to
        generate the secrets, but now both partners can. At a minimum you
        would need to share only the account name and the user name if one
        is needed. With that, the other partner can generate the passcode.

        Once you have shared an accounts file, you can also use the identity
        command to prove your identity to your partner.

        You cannot share secrets encrypted with Scrypt. Also, you cannot
        share stealth accounts unless the file that contains the account
        templates has a master_seed specified, which they do not by
        default. You would need to create a separate file for shared stealth
        account templates and add a master seed to that file manually.
    """).strip()
    assert result.decode('utf-8') == expected

# test_conceal() {{{1
def test_conceal():
    try:
        result = run('avendesora help conceal')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Conceal text by encoding it.

        Usage:
            avendesora conceal [options] [<text>]
            avendesora c       [options] [<text>]

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when concealing information.
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
                                    Use commas with no spaces to separate multiple IDs.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).
            -s, --symmetric         Encrypt with a passphrase rather than using your
                                    GPG key (only appropriate for gpg encodings).

        Possible encodings include (default encoding is base64):

        gpg:
            This encoding fully encrypts/decrypts the text with GPG key.
            By default your GPG key is used, but you can specify symmetric
            encryption, in which case a passphrase is used.

        base64:
            This encoding obscures but does not encrypt the text. It can
            protect text from observers that get a quick glance of the
            encoded text, but if they are able to capture it they can easily
            decode it.

        scrypt:
            This encoding fully encrypts the text with your user key. Only
            you can decrypt it, secrets encoded with scrypt cannot be shared.

        Though available as an option for convenience, you should not pass
        the text to be hidden as an argument as it is possible for others to
        examine the commands you run and their argument list. For any
        sensitive secret, you should simply run 'avendesora conceal' and
        then enter the secret text when prompted.
    """).strip()
    assert result.decode('utf-8') == expected

# test_credentials() {{{1
def test_credentials():
    try:
        result = run('avendesora help credentials')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Show login credentials.

        Displays the account's login credentials, which generally consist of an
        identifier and a secret.

        Usage:
            avendesora credentials [options] <account>
            avendesora login       [options] <account>
            avendesora l           [options] <account>

        Options:
            -S, --seed              Interactively request additional seed for
                                    generated secrets.

        The credentials can be specified explicitly using the credentials
        setting in your account. For example::

            credentials = 'usernames.0 usernames.1 passcode'

        If credentials is not specified then the first of the following will
        be used if available:

            id: username or email
            secret: passcode, password or passphrase

        If your credentials include more than one secret they will be
        presented one at a time for one minute each. You can cut the minute
        short by typing Ctrl-C.
    """).strip()
    assert result.decode('utf-8') == expected

# test_edit() {{{1
def test_edit():
    try:
        result = run('avendesora help edit')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Edit an account.

        Usage:
            avendesora edit <account>
            avendesora e    <account>

        Opens an existing account in your editor.

        You can specify the editor by changing the 'edit_account' setting in
        the config file (~/.config/avendesora/config).
    """).strip()
    assert result.decode('utf-8') == expected

# test_find() {{{1
def test_find():
    try:
        result = run('avendesora help find')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Find an account.

        Find accounts whose name contains the search text.

        Usage:
            avendesora find <text>
            avendesora f    <text>
    """).strip()
    assert result.decode('utf-8') == expected

# test_help() {{{1
def test_help():
    try:
        result = run('avendesora help help')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Give information about commands or other topics.

        Usage:
            avendesora help [options] [<topic>]
            avendesora h    [options] [<topic>]

        Options:
            -s, --search            list topics that include <topic> as a search term.
            -b, --browse            open the topic in your default browser.
    """).strip()
    assert result.decode('utf-8') == expected

# test_identity() {{{1
def test_identity():
    try:
        result = run('avendesora help identity')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Generate an identifying response to a challenge.

        Usage:
            avendesora identity [<name> [<challenge>...]]
            avendesora ident    [<name> [<challenge>...]]
            avendesora i        [<name> [<challenge>...]]

        This command allows you to generate a response to any challenge.
        The response identifies you to a partner with whom you have shared
        an account.

        If you run the command with no arguments, it prints the list of
        valid names. If you run it with no challenge, one is created for you
        based on the current time and date.

        If you have a remote partner to whom you wish to prove your
        identity, have that partner use avendesora to generate a challenge
        and a response based on your shared secret. Then the remote partner
        provides you with the challenge and you run avendesora with that
        challenge to generate the same response, which you provide to your
        remote partner to prove your identity.

        You are free to explicitly specify a challenge to start the process,
        but it is important that it be unpredictable and that you not use
        the same challenge twice. As such, it is recommended that you not
        provide the challenge. In this situation, one is generated for you
        based on the time and date.

        Consider an example that illustrates the process. In this example,
        Siuan is confirming the identity of Moiraine, where both Siuan and
        Moiraine are assumed to have shared *Avendesora* accounts.  Siuan
        runs *Avendesora* as follows and remembers the response::

            > avendesora identity moiraine
            challenge: slouch emirate bedeck brooding
            response: spear disable local marigold

        This assumes that moiraine is the name, with any extension removed,
        of the file that Siuan uses to contain their shared accounts.

        Siuan communicates the challenge to Moiraine but not the response.
        Moiraine then runs *Avendesora* with the given challenge::

            > avendesora identity siuan slouch emirate bedeck brooding
            challenge: slouch emirate bedeck brooding
            response: spear disable local marigold

        In this example, siuan is the name of the file that Moiraine uses to
        contain their shared accounts.

        To complete the process, Moiraine returns the response to Siuan, who
        compares it to the response she received to confirm Moiraine's
        identity.  If Siuan has forgotten the desired response, she can also
        specify the challenge to the :ref:`identity command <identity
        command>` to regenerate the expected response.

        Alternately, when Siuan sends a message to Moiraine, she can
        proactively prove her identity by providing both the challenge and
        the response. Moiraine could then run the *identity* command with
        the challenge and confirm that she gets the same response. Other
        than herself, only Siuan could predict the correct response to any
        challenge.  However, this is not recommended as it would allow
        someone with brief access to Suian's Avendesora, perhaps Leane her
        Keeper, to generate and store multiple challenge/response pairs.
        Leane could then send messages to Moiraine while pretending to be
        Siuan using the saved challenge/response pairs.  The subterfuge
        would not work if Moiraine generated the challenge unless Leane
        currently has access to Siuan's Avendesora.
    """).strip()
    assert result.decode('utf-8') == expected

# test_initialize() {{{1
def test_initialize():
    try:
        result = run('avendesora help initialize')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Create initial set of Avendesora files.

        Usage:
            avendesora initialize [options]
            avendesora init       [options]
            avendesora I          [options]

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
                                    Use commas with no spaces to separate multiple IDs.
            -h <path>, --gpg-home <path>
                                    GPG home directory (default is ~/.gnupg).

        Create Avendesora data directory (~/.config/avendesora) and populate
        it with initial versions of all essential files.

        It is safe to run this command even after the data directory and
        files have been created. Doing so will simply recreate any missing
        files.  Existing files are not modified.
    """).strip()
    assert result.decode('utf-8') == expected

# test_log() {{{1
def test_log():
    try:
        result = run('avendesora help log')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Open the logfile.

        Usage:
            avendesora [options] log

        Options:
            -d, --delete   Delete the logfile rather than opening it.

        Opens the logfile in your editor.

        You can specify the editor by changing the 'edit_account' setting in
        the config file (~/.config/avendesora/config).
    """).strip()
    assert result.decode('utf-8') == expected

# test_new() {{{1
def test_new():
    try:
        result = run('avendesora help new')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Create new accounts file.

        Usage:
            avendesora new [options] <name>
            avendesora N   [options] <name>

        Options:
            -g <id>, --gpg-id <id>  Use this ID when creating any missing encrypted files.
                                    Use commas with no spaces to separate multiple IDs.

        Creates a new accounts file. Accounts that share the same file share
        the same master seed by default and, if the file is encrypted,
        can be decrypted by the same recipients.

        Generally you create new accounts files for each person or group
        with which you wish to share accounts. You also use separate files
        for passwords with different security domains. For example, a
        high-value passwords might be placed in an encrypted file that would
        only be placed highly on protected computers. Conversely, low-value
        passwords might be contained in perhaps an unencrypted file that is
        found on many computers.

        Add a '.gpg' extension to <name> to encrypt the file.
    """).strip()
    assert result.decode('utf-8') == expected

# test_phonetic() {{{1
def test_phonetic():
    try:
        result = run('avendesora help phonetic')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display NATO phonetic alphabet.

        Usage:
            avendesora alphabet [<text>]
            avendesora phonetic [<text>]
            avendesora p [<text>]

        If <text> is given, it is converted character by character to the
        phonetic alphabet. If not given, the entire phonetic alphabet is
        displayed.
    """).strip()
    assert result.decode('utf-8') == expected

# test_questions() {{{1
def test_questions():
    try:
        result = run('avendesora help questions')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Answer a security question.

        Displays the security questions and then allows you to select one
        to be answered.

        Usage:
            avendesora questions [options] <account> [<field>]
            avendesora quest     [options] <account> [<field>]
            avendesora q         [options] <account> [<field>]
            avendesora qc        [options] <account> [<field>]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -S, --seed              Interactively request additional seed for
                                    generated secrets.

        The 'qc' command is a shortcut for 'questions --clipboard'.

        Request the answer to a security question by giving the account name to
        this command.  For example:

            avendesora questions bank

        It will print out the security questions for *bank* account along with
        an index. Specify the index of the question you want answered.

        By default *Avendesora* looks for the security questions in the
        *questions* field.  If your questions are in a different field, just
        specify the name of the field on the command line:

            avendesora questions bank verbal
    """).strip()
    assert result.decode('utf-8') == expected

# test_reveal() {{{1
def test_reveal():
    try:
        result = run('avendesora help reveal')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Reveal concealed text.

        Transform concealed text to reveal its original form.

        Usage:
            avendesora reveal [<text>]
            avendesora r      [<text>]

        Options:
            -e <encoding>, --encoding <encoding>
                                    Encoding used when revealing information.

        Though available as an option for convenience, you should not pass
        the text to be revealed as an argument as it is possible for others
        to examine the commands you run and their argument list. For any
        sensitive secret, you should simply run 'avendesora reveal' and then
        enter the encoded text when prompted.
    """).strip()
    assert result.decode('utf-8') == expected

# test_search() {{{1
def test_search():
    try:
        result = run('avendesora help search')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Search accounts.

        Search for accounts whose values contain the search text.

        Usage:
            avendesora search <text>
            avendesora s      <text>
    """).strip()
    assert result.decode('utf-8') == expected

# test_value() {{{1
def test_value():
    try:
        result = run('avendesora help value')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Show an account value.

        Produce an account value. If the value is secret, it is produced only
        temporarily unless --stdout is specified.

        Usage:
            avendesora value [options] [<account> [<field>]]
            avendesora val   [options] [<account> [<field>]]
            avendesora v     [options] [<account> [<field>]]
            avendesora vc    [options] [<account> [<field>]]

        Options:
            -c, --clipboard         Write output to clipboard rather than stdout.
            -s, --stdout            Write output to the standard output without
                                    any annotation or protections.
            -S, --seed              Interactively request additional seed for
                                    generated secrets.
            -v, --verbose           Add additional information to log file to
                                    help identify issues in account discovery.
            -T <title>, --title <title>
                                    Use account discovery on this title.

        The 'vc' command is a shortcut for 'value --clipboard'.

        You request a scalar value by specifying its name after the account.
        For example:

            avendesora value bank pin

        If the requested value is composite (an array or dictionary), you should
        also specify a key that indicates which of the composite values you
        want. For example, if the 'accounts' field is a dictionary, you specify
        accounts.checking or accounts[checking] to get information on your
        checking account. If the value is an array, you give the index of the
        desired value. For example, questions.0 or questions[0]. If you only
        specify a number, then the name is assumed to be 'questions', as in the
        list of security questions (this can be changed by specifying the
        desired name as the 'default_vector_field' in the account or the config
        file).

        The field may be also be a script, with is nothing but a string that it
        output as given except that embedded attributes are replaced by account
        field values. For example:

            avendesora value bank '{accounts.checking}: {passcode}'

        If no value is requested the result produced is determined by the value
        of the 'default' attribute (this can be changed by specifying
        'default_field' in the config file). If no value is given for 'default',
        then the *passcode*, *password*, or *passphrase* attribute is produced
        (this can be changed by specifying the 'default_field' in the account or
        the config file).  If 'default' is a script (see 'avendesora help
        scripts') then the script is executed. A typical script might be
        'username: {username}, password: {passcode}'. It is best if the script
        produces a one line output if it contains secrets. If not a script, the
        value of 'default' should be the name of another attribute, and the
        value of that attribute is shown.

        Normally the value command attempts to protects secrets. It does so
        clearing the screen after a minute. If multiple secrets are requested,
        you must either wait a minute to see each subsequent secret or type
        Ctrl-C to clear the current secret and move on.  If you use --clipboard,
        the clipboard is cleared after a minute.  However, if you use --stdout
        this clearing of the secret does not occur. The --stdout option is
        generally used with communicating with other Linux commands.  For
        example, you can send a passcode to the standard input of a command as
        follows:

            avendesora value --stdout gpg | gpg --passphrase-fd 0 ...

        You can place the username and password on a command line as follows:

            curl --user `avendesora value -s apache '{username}:{passcode}'` ...

        Be aware that it is possible for other users on shared Linux machines to
        see the command line arguments of your commands, so passing secrets as
        command arguments should only be used for low value secrets.

        If no account is requested, then Avendesora attempts to determine the
        appropriate account through discovery (see 'avendesora help discovery').
        Normally Avendesora is called in this manner from your window manager.
        You would arrange for it to be run when you type a hot key. In this case
        Avendesora determines which account to use from information available
        from the environment, information like the title on active window. In
        this mode, Avendesora mimics the keyboard when producing its output.

        The verbose and title options are used when debugging account
        discovery. The verbose option adds more information about the
        discovery process to the logfile (~/.config/avendesora/log.gpg). The
        title option allows you to override the active window title so you can
        debug title-based discovery. Specifying the title option also scrubs
        the output and outputs directly to the standard output rather than
        mimicking the keyboard so as to avoid exposing your secret.

        If the --stdout option is not specified, the value command still writes
        to the standard output if it is associated with a TTY (if Avendesora is
        outputting directly to a terminal). If standard output is not a TTY,
        Avendesora mimics the keyboard and types the desired value directly into
        the active window.  There are two common situations where standard
        output is not a TTY: when Avendesora is being run by your window manager
        in response to you pressing a hot key or when the output of Avendesora
        is fed into a pipeline.  In the second case, mimicking the keyboard is
        not what you want; you should use --stdout to assure the chosen value is
        sent to the pipeline as desired.
    """).strip()
    assert result.decode('utf-8') == expected

# test_values() {{{1
def test_values():
    try:
        result = run('avendesora help values')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display all account values.

        Show all account values.

        Usage:
            avendesora values <account>
            avendesora vals   <account>
            avendesora V      <account>
    """).strip()
    assert result.decode('utf-8') == expected

# test_version() {{{1
def test_version():
    try:
        result = run('avendesora help version')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Display Avendesora version.

        Usage:
            avendesora version
    """).strip()
    assert result.decode('utf-8') == expected

# test_abraxas() {{{1
def test_abraxas():
    try:
        result = run('avendesora help abraxas')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Avendesora generalizes and replaces Abraxas, its predecessor.  To
        transition from Abraxas to Avendesora, you will first need to
        upgrade Abraxas to version 1.8 or higher (use 'abraxas -v' to
        determine version). Then run:

            abraxas --export

        It will create a collection of Avendesora accounts files in
        ~/.config/abraxas/avendesora. You need to manually add these files
        to your list of accounts files in Avendesora. Say one such file is
        created: ~/.config/abraxas/avendesora/accounts.gpg.  This could be
        added to Avendesora as follows:

        1. create a symbolic link from
           ~/.config/avendesora/abraxas_accounts.gpg to
           ~/.config/abraxas/avendesora/accounts.gpg:

            cd ~/.config/avendesora
            ln -s ../abraxas/avendesora/accounts.gpg abraxas_accounts.gpg

        2. add abraxas_accounts.gpg to account_files list in accounts_files.

        Now all of the Abraxas accounts contained in abraxas_accounts.gpg
        should be available though Avendesora and the various features of
        the account should operate as expected. However, secrets in accounts
        exported by Abraxas are no longer generated secrets. Instead, the
        actual secrets are placed in a hidden form in the exported accounts
        files.

        If you would like to enhance the imported accounts to take advantage
        of the new features of Avendesora, it is recommended that you do not
        manually modify the imported files. Instead, copy the account
        information to one of your own account files before modifying it.
        To avoid conflict, you must then delete the account from the
        imported file. To do so, create ~/.config/abraxas/do-not-export if
        it does not exist, then add the account name to this file, and
        reexport your accounts from Abraxas.
    """).strip()
    assert result.decode('utf-8') == expected

# test_accounts() {{{1
def test_accounts():
    try:
        result = run('avendesora help accounts')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Account information is stored in account files. The list of account
        files is given in ~/.config/avendesora/accounts_files.  New account
        files are created using 'avendesora new', but to delete an accounts
        file, you must remove it from accounts_files. Once an accounts file
        exists, you may add accounts to it using 'account add'. Use the -f
        option to specify which file is to contain the new account.

        An account is basically a collection of attributes organized as a
        subclass of the Python Account class. For example:

            class NewYorkTimes(Account):
                aliases = 'nyt'
                username = 'derrickAsh'
                email = 'derrickAsh@gmail.com'
                passcode = PasswordRecipe('12 2u 2d 2s')
                discovery = RecognizeURL(
                    'https://myaccount.nytimes.com',
                    script='{email}{tab}{passcode}{return}'
                )

        Most of the field values can be retrieved simply by asking for them.
        For example,

            > avendesora value newyorktimes username
            username: derrickAsh

        In general, values can be strings, arrays, dictionaries, and special
        Advendesora classes. For example, you could have an array of
        security questions:

            questions = [
                Question("What is your mother's maiden name?"),
                Question("What city were you born?"),
                Question("What is first pet's name?"),
            ]

        Then you can request the answer to a particular question using its
        index:

            > avendesora value newyorktimes questions.0
            questions.0 (What is your mother's maiden name?): portrayal tentacle fanlight

        'questions' is the default array field, so you could have shortened
        your request by using '0' rather than 'questions.0'.  You might be
        thinking, hey, that is not my mother's maiden name. That is because
        Question is a 'generated secret'.  It produces a completely arbitrary
        answer that is impossible to predict. Thus, even family members
        cannot know the answers to your security questions.

        A dictionary is often used to hold account numbers:

            class Bank(Account):
                accounts = {
                    'checking': '1234-56-7890',
                    'savings': '0123-45-6789',
                }

        You then access its values using:

            > avendesora value bank accounts.checking
            accounts.checking: 1234-56-7890

        You might consider your account numbers as sensitive information. In
        this case you can hide them with:

            class Bank(Account):
                accounts = {
                    'checking': Hidden('MTIzNC01Ni03ODkw'),
                    'savings': Hidden('MDEyMy00NS02Nzg5'),
                }

        The values are now hidden, but not encrypted. They are simply
        encoded with base64. Any knowledgeable person with the encoded value
        can decode it back to its original value. Using Hidden makes it
        harder to recognize and remember the value given only a quick
        over-the-shoulder glance. It also marks the value as sensitive, so
        it will only be displayed for a minute. You generate the encoded
        value using 'avendesora conceal'.

        You can find the specifics of how to specify or generate your
        secrets by running 'avendesora help secrets'.

        Any value that is an instance of the GeneratedSecret class (Password,
        Passphrase, ...) or the ObscuredSecret class (Hidden, GPG, ...) are
        considered sensitive. They are given out only in a controlled
        manner. For example, running 'avendesora values' displays all
        fields, but the values that are sensitive are replaced by
        instructions on how to view them. They can only be viewed
        individually:

            > avendesora values newyorktimes
            names: newyorktimes, nyt
            email: derrickAsh@gmail.com
            passcode: <reveal with 'avendesora value newyorktimes passcode'>
            username: derrickAsh

        The 'aliases' and 'discovery' fields are not shown because they are
        considered tool fields. Other tool fields include 'NAME', 'default',
        'master', 'browser', and 'default_url'. For more information on
        discovery, run 'avendesora help discovery'.  'default' is the name
        of the default field, which is the field you get if you do not
        request a particular field. Its value defaults to 'passcode' but it
        can be set to any account attribute name or it can be a script (see
        'avendesora help scripts').  'browser' is the default browser to use
        when opening the account, run 'avendesora help browse' to see a list
        of available browsers.

        The value of passcode is considered sensitive because it is an
        instance of PasswordRecipe, which is a subclass of GeneratedSecret.
        If we wish to see the passcode, use:

            > avendesora value nyt
            passcode: TZuk8:u7qY8%

        This value will be displayed for a minute and then hidden. If you
        would like to hide it early, simply type Ctrl-C.

        An attribute value can incorporate other attribute values through
        use of the Script class. For example, consider an account for your
        wireless router that contains the following:

            ssid = {
                'huron_guests': Passphrase(),
                'huron_drugs': Passphrase(),
            }
            guest = Script('SSID: huron_guests, password: {ssid.huron_guests}')
            privileged = Script('SSID: huron_drugs, password: {ssid.huron_drugs}')

        The ssid field is a dictionary that contains the SSID and pass
        phrases for each of the wireless networks provided by the router.
        This is a natural an compact representation for this information,
        but accessing it as a user in this form would require two steps to
        access the information, one to get the SSID and another to get the
        passphrase. This issue is addressed by adding the guest and
        privileged attributes. The guest and privileged attributes are a
        script that gives the SSID and interpolate the pass phrase. Now both
        can easily accessed at once with:

            > avendesora value wifi guest
            SSID: huron_guests, password: delimit ballcock fibber levitate

        Use of Avendesora classes (GeneratedSecret, ObscuredSecret, or
        Script) is confined to the top two levels of account attributes,
        meaning that they can be the value of the top-level attributes, or
        the top-level attributes may be arrays or dictionaries that contain
        objects of these classes, but it can go no further.

        By default the account name is taken to be the class name converted
        to lower case.  So the name for the NewYorkTimes account given above
        would be newyorktimes. You can override the name with the NAME
        field. This allows you to create account names that contain dashes
        or any other character that would not be acceptable in a class name
        (best to avoid characters that are meaningful to your shell, such as
        !$&*()|<>'"\{};`.  For example:

            class Root_Work(Account):
                NAME = 'root@work'

        It is important to remember that any generated secrets use the
        account name when generating their value, so if you change the
        account name you will change the secret.  For this reason is it
        important to choose a good account name up front and not change it.
        It should be very specific to avoid conflicts with similar accounts
        created later.  For example, rather than choosing gmail as your
        account name, you might want to include your username, ex.
        gmail-paul-bunyan. This would allow you to create additional gmail
        accounts later without ambiguity. Then just add 'gmail' as an alias
        to the account you use most often.
    """).strip()
    assert result.decode('utf-8') == expected

# test_browsing() {{{1
def test_browsing():
    try:
        result = run('avendesora help browsing')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Avendesora provides the browse command to allow you to easily open
        the website for your account in your browser. To do so, it needs two
        things: a URL and a browser.

        Avendesora looks for URLs in the urls and discovery account
        attributes, with urls being preferred if both exist.  urls may
        either be a string, a list, or a dictionary. If it is a string, it
        is split at white spaces to make it a list.  If urls is a list, the
        URLs are considered unnamed and the first one given is used. If it a
        dictionary, the URLs are named.  When named, you may specify the URL
        you wish to use by specifying the name to the browse command.  For
        example, consider a urls attribute that looks like this:

            class Dragon(Account):
                username = 'rand'
                passcode = Passphrase()
                urls = dict(
                    email = 'https://webmail.dragon.com',
                    vpn = 'https://vpn.dragon.com',
                )
                default_url = 'email'

        You would access vpn with:

            avendesora browse dragon vpn

        By specifying default_url you indicate which URL is desired when you
        do not explicitly specify which you want on the browse command. In
        this way, you can access your email with either of the following:

            avendesora browse dragon email
            avendesora browse dragon

        If urls is not given, Avendesora looks for URLs in RecognizeURL
        members in the discovery attribute.  If the name argument is
        provided to RecognizeURL, it is treated as a named URL, otherwise it
        is treated as an unnamed URL.

        If named URLs are found in both urls and discovery they are all
        available to browse command, with those given in *urls* being
        preferred when the same name is found in both attributes.

        You can configure browsers for use by Avendesora using the browsers
        setting.  By default, browsers contains the following:

            browsers = dict(
                f = 'firefox -new-tab {url}',
                fp = 'firefox -private-window {url}',
                c = 'google-chrome {url}',
                ci = 'google-chrome --incognito {url}',
                q =  'qutebrowser {url}',
                t = 'torbrowser {url}',
                x = 'xdg-open {url}',
            )

        Each entry pairs a key with a command. The command will be run with {url} 
        replaced by the selected URL when the browser is selected. You can choose which 
        browser is used by specifying the --browser command line option on the 
        browse command, by adding the browser attribute to the account, or by 
        specifying the default_browser setting in the :ref:`config file 
        <configuring>`.  If more than one is specified, the command line option 
        dominates over the account attribute, which dominates over the setting.  By 
        default, the default browser is x, which uses the default browser for your 
        account.
    """).strip()
    assert result.decode('utf-8') == expected

# test_collaborate() {{{1
def test_collaborate():
    try:
        result = run('avendesora help collaborate')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        If you share an accounts file with a partner, then either partner
        can create new secrets and the other partner can reproduce their
        values once a small amount of relatively non-confidential
        information is shared. This works because the security of the
        generated secrets is based on the master seed, and that seed is
        contained in the accounts file that is shared in a secure manner
        once at the beginning.  For example, imagine one partner creates an
        account at the US Postal Service website and then informs the
        partner that the name of the new account is usps and the username is
        justus.  That is enough information for the second partner to
        generate the password and login. And notice that the necessary
        information can be shared over an insecure channel. For example, it
        could be sent in a text message or from a phone where trustworthy
        encryption is not available.

        The first step in using Avendesora to collaborate with a partner is
        for one of the partners to generate and then share an accounts file
        that is dedicated to the shared accounts.  This file contains the
        master seed, and it is critical to keep this value secure. Thus, it
        is recommended that the shared file be encrypted.

        Consider an example where you, Alice, are sharing accounts with your
        business partner, Bob.  You have hired a contractor to run your
        email server, Eve, who unbeknownst to you is reading your email in
        order to steal valuable secrets.  Together, you and Bob jointly run
        Teneya Enterprises. Since you expect more people will need access to
        the accounts in the future, you choose to the name the file after
        the company rather than your partner.  To share accounts with Bob,
        you start by getting Bob's public GPG key.  Then, create the new
        accounts file with something like:

            avendesora new -g alice@teneya.com -g bob@teneya.com teneya.gpg

        This generates a new accounts file, ~/.config/avendesora/teneya.gpg,
        and encrypts it so only you and Bob can open it.  Mail this file to
        Bob. Since it is encrypted, it is to safe to send the file through
        email.  Even though Eve can read this message, the accounts file is
        encrypted so Eve cannot access the master seed it contains.  Bob
        should put the file in ~/.config/avendesora and then add it to
        accounts_files in ~/.config/avendesora/accounts_files.  You are now
        ready to share accounts.

        Then, one partner creates a new account and mails the account entry
        to the other partner.  This entry does not contain enough
        information to allow an eavesdropper such as Eve to be able to
        generate the secrets, but now both partners can. At a minimum you
        would need to share only the account name and the user name if one
        is needed. With that, the other partner can generate the passcode.

        Once you have shared an accounts file, you can also use the identity
        command to prove your identity to your partner.

        You cannot share secrets encrypted with Scrypt. Also, you cannot
        share stealth accounts unless the file that contains the account
        templates has a master_seed specified, which they do not by
        default. You would need to create a separate file for shared stealth
        account templates and add a master seed to that file manually.
    """).strip()
    assert result.decode('utf-8') == expected

# test_discovery() {{{1
def test_discovery():
    try:
        result = run('avendesora help discovery')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        If you do not give an account to 'avendesora value', Avendesora will
        try to determine the account by simply asking each account. An
        account can look at the window title, the user name, the host name,
        the working directory, and the environment variables to determine if
        it is suitable. If so, it nominates itself. If there is only one
        account nominated, that account is used. If there are multiple
        nominees, then a small window pops up allowing you to choose
        which account you wish to use.

        To configure an account to trigger when a particular window title is
        seen, use:

            discovery = RecognizeTitle(
                'Chase Online *',
                script='{username}{tab}{passcode}{return}'
            )

        The title is a glob string, meaning that '*' matches any combination
        of characters. The script describes what Avendesora should output
        when their is a match. In this case it outputs the username field,
        then a tab, then the passcode field, then a return (see 'avendesora
        help scripts').

        Matching window titles can be fragile, especially for websites
        because the titles can vary quite a bit across the site and over
        time. To accommodate this variation, you can give multiple glob
        patterns:

            discovery = RecognizeTitle(
                'CHASE Bank*',
                'Chase Online*',
                script='{username}{tab}{passcode}{return}'
            )

        If you use Firefox, you can install the 'Add URL to Window Title'
        extension.  It is a plugin that makes discovery easier and more
        robust by adding the URL to the title.  For Chrome the appropriate
        plugin is  is 'URL in Title'.  It is recommended that you install
        the appropriate one into your browser.  For AddURLToWindowTitle, set
        the following options:

            show full URL = yes
            separator string = '-'
            show field attributes = no

        For URLinTitle, set:

            tab title format = '{title} - {protocol}://{hostname}{port}/{path}'

        RecognizeURL designed to recognize such titles. Once you have
        deployed the appropriate plugin, you can use:

            discovery = RecognizeURL(
                'https://chaseonline.chase.com',
                'https://www.chase.com',
                script='{username}{tab}{passcode}{return}'
            )

        When giving the URL, anything specified must match and globbing is
        not supported. If you give a partial path, by default Avendesora
        will match up to what you have given, but you can require an exact
        match of the entire path by specifying exact_path=True to
        RecognizeURL.  If you do not give the protocol, the default_protocol
        (https) is assumed.

        In general you should use RecognizeURL() rather than
        RecognizeTitle() for websites if you can. Doing so will help protect
        you from phishing attacks by carefully examining the URL.

        When account discovery fails it can be difficult to determine what
        is going wrong. When this occurs, you should first examine the log
        file. It should show you the window title and the recognized title
        components. You should first assure the title is as expected. If Add
        URL to Window Title generated the title, then the various title
        components should also be shown.  Then run Avendesora as follows:

            avendesora value --verbose --title '<title>'

        The title should be copied from the log file. The verbose option
        causes the result of each test to be included in the log file, so
        you can determine which recognizer is failing to trigger.  You can
        either specify the verbose option on the command line or in the
        config file.

        The following recognizers are available:

            RecognizeAll(<recognizer>..., [script=<script>])
            RecognizeAny(<recognizer>..., [script=<script>])
            RecognizeTitle(<title>..., [script=<script>])
            RecognizeURL(<title>..., [script=<script>, [name=<name>,]] [exact_path=<bool>])
            RecognizeHost(<host>..., [script=<script>])
            RecognizeUser(<user>..., [script=<script>])
            RecognizeCWD(<cwd>..., [script=<script>])
            RecognizeEnvVar(<name>, <value>, [script=<script>])
            RecognizeNetwork(<mac>..., [script=<script>])
            RecognizeFile(<path>, [<contents>,] [<ttl>,] [script=<script>])

        RecognizeAll() and RecognizeAny() can be used to combine several
        recognizers. For example:

            discovery = RecognizeAll(
                RecognizeTitle('sudo *'),
                RecognizeUser('hhyde'),
                script='{passcode}{return}'
            )

        If the recognizers are given in an array, all are tried, and each
        that match are offered. For example:

            discovery = [
                RecognizeURL(
                    'http://www.querty-forum.org',
                    script='admin{tab}{passcode}{return}',
                    name='admin',
                ),
                RecognizeURL(
                    'http://www.querty-forum.org',
                    script='thecaretaker{tab}{passcode}{return}',
                    name='thecaretaker',
                ),
            ]

        In this case, both recognizers recognize the same URL, thus they
        will both be offered for this site.  But each has a different
        script. The name allows the user to distinguish the available
        choices.

        If there is a need to distinguish URLs where one is a substring
        of another, you can use *exact_path*::

            discovery = [
                RecognizeURL(
                    'https://mybank.com/Authentication',
                    script='{username}{return}',
                    exact_path=True,
                ),
                RecognizeURL(
                    'https://mybank.com/Authentication/Password',
                    script='{passcode}{return}',
                    exact_path=True,
                ),
            ]

        Some similar URLs can be distinguished by their fragment. If the
        URL contains '#', then anything that follows is considered its
        fragment.  For example::

            discovery = [
                RecognizeURL(
                    'https://mybank.com/auth',
                    script='{username}{tab}{passcode}{return}',
                    fragment='/',
                ),
                RecognizeURL(
                    'https://mybank.com/auth',
                    script='{passcode}{return}',
                    fragment='/password',
                ),
            ]

        In this example, the URL 'https://mybank.com/auth/#/' would match
        the first entry, which outputs both the username and password, and a
        URL of 'https://mybank.com/auth/#/password' would match the second,
        which only outputs the password.

        RecognizeFile checks to determine whether a particular file has been
        created recently.  This can be use in scripts to force secret
        recognition.  For example, the titles used by Firefox and
        Thunderbird when collecting the master password is either
        non-existent or undistinguished.  These programs also produce a
        large amount of uninteresting chatter on their output, so it is
        common to write a shell script to run the program that redirects
        their output to /dev/null.  Such a script can be modified to
        essentially notify Avendesora that a particular password is desired.
        For example, for Thunderbird:

            #!/bin/sh
            touch /tmp/thunderbird-1024
            /usr/bin/thunderbird > /dev/null

        Here I have adding my user id (uid=1024) to make the filename unique
        so I am less likely to clash with other users. Alternately, you
        could choose a path that fell within your home directory. Then,
        adding:

            class Firefox(Account):
                desc = 'Master password for Firefox and Thunderbird'
                passcode = Password()
                discovery = RecognizeFile(
                    '/tmp/thunderbird-1024', wait=60, script='{passcode}{return}'
                )

        If the specified file exists and has been updated within the last 60
        seconds, then secret is recognized.  You can specify the amount of
        time you can wait in between running the script and running
        Avendesora with the 'wait' argument, which takes a number of
        seconds.  It defaults to 60.

        Using this particular approach, every secret would need its own
        file. But you can share a file by specifying the file contents.
        Then the script could be rewritten as:

            #!/bin/sh
            echo thunderbird > ~/.avendesora-password-request
            /usr/bin/thunderbird > /dev/null

        Then you would add something like the following to your accounts file:

            class Firefox(Account):
                desc = 'Master password for Firefox and Thunderbird'
                passcode = Password()
                discovery = RecognizeFile(
                    '~/.avendesora-password-request',
                    contents='thunderbird',
                    script='{passcode}{return}'
                )
    """).strip()
    assert result.decode('utf-8') == expected

# test_entropy() {{{1
def test_entropy():
    try:
        result = run('avendesora help entropy')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        A 4 word Avendesora password provides 53 bits of entropy, which
        seems like a lot, but NIST is recommending 80 bits for your most
        secure passwords.  So, how much is actually required. It is worth
        exploring this question.  Entropy is a measure of how hard the
        password is to guess. Specifically, it is the base two logarithm of
        the likelihood of guessing the password in a single guess. Every
        increase by one in the entropy represents a doubling in the
        difficulty of guessing your password. The actual entropy is hard to
        pin down, so generally we talk about the minimum entropy, which is
        the likelihood of an adversary guessing the password if he or she
        knows everything about the scheme used to generate the password but
        does not know the password itself.  So in this case the minimum
        entropy is the likelihood of guessing the password if it is known
        that we are using 4 space separated lower case words as our pass
        phrase.  This is very easy to compute.  There are roughly 10,000
        words in our dictionary, so if there was only one word in our pass
        phrase, the chance of guessing it would be one in 10,000 or 13 bits
        of entropy.  If we used a two word pass phrase the chance of
        guessing it in a single guess is one in 10,000*10,000 or one in
        100,000,000 or 26 bits of entropy.

        The probability of guessing our pass phrase in one guess is not our
        primary concern. Really what we need to worry about is given a
        determined attack, how long would it take to guess the password. To
        calculate that, we need to know how fast our adversary could try
        guesses. If they are trying guesses by typing them in by hand, their
        rate is so low, say one every 10 seconds, that even a one word pass
        phrase may be enough to deter them.  Alternatively, they may have a
        script that automatically tries pass phrases through a login
        interface.  Again, generally the rate is relatively slow.  Perhaps
        at most the can get is 1000 tries per second. In this case they
        would be able to guess a one word pass phrase in 10 seconds and a
        two word pass phrase in a day, but a 4 word pass phrase would
        require 300,000 years to guess in this way.

        The next important thing to think about is how your password is
        stored by the machine or service you are logging into. The worst
        case situation is if they save the passwords in plain text. In this
        case if someone were able to break in to the machine or service,
        they could steal the passwords. Saving passwords in plain text is an
        extremely poor practice that was surprisingly common, but is
        becoming less common as companies start to realize their liability
        when their password files get stolen.  Instead, they are moving to
        saving passwords as hashes.  A hash is a transformation that is very
        difficult to reverse, meaning that if you have the password it is
        easy to compute its hash, but given the hash it is extremely
        difficult to compute the original password. Thus, they save the
        hashes (the transformed passwords) rather than the passwords. When
        you log in and provide your password, it is transformed with the
        hash and the result is compared against the saved hash. If they are
        the same, you are allowed in. In that way, your password is no
        longer available to thieves that break in.  However, they can still
        steal the file of hashed passwords, which is not as good as getting
        the plain text passwords, but it is still valuable because it allows
        thieves to greatly increase the rate that they can try passwords. If
        a poor hash was used to hash the passwords, then passwords can be
        tried at a very high rate.  For example, it was recently reported
        that password crackers were able to try 8 billion passwords per
        second when passwords were hashed with the MD5 algorithm. This would
        allow a 4 word pass phrase to be broken in 14 days, whereas a 6 word
        password would still require 4,000,000 years to break.  The rate for
        the more computational intensive sha512 hash was only 2,000
        passwords per second. In this case, a 4 word pass phrase would
        require 160,000 years to break.

        In most cases you have no control over how your passwords are stored
        on the machines or services that you log into.  Your best defense
        against the notoriously poor security practices of most sites is to
        always use a unique password for sites where you are not in control
        of the secrets.  For example, you might consider using the same pass
        phrase for you login password and the pass phrase for an ssh key on
        a machine that you administer, but never use the same password for
        two different websites unless you do not care if the content of
        those sites become public.

        So, if we return to the question of how much entropy is enough, you
        can say that for important passwords where you are in control of the
        password database and it is extremely unlikely to get stolen, then
        four randomly chosen words from a reasonably large dictionary is
        plenty.  If what the pass phrase is trying to protect is very
        valuable and you do not control the password database (ex., your
        brokerage account) you might want to follow the NIST recommendation
        and use 6 words to get 80 bits of entropy. If you are typing
        passwords on your work machine, many of which employ keyloggers to
        record your every keystroke, then no amount of entropy will protect
        you from anyone that has or gains access to the output of the
        keylogger.  In this case, you should consider things like one-time
        passwords or two-factor authentication. Or better yet, only access
        sensitive accounts from your home machine and not from any machine
        that you do not control.
    """).strip()
    assert result.decode('utf-8') == expected

# test_misdirection() {{{1
def test_misdirection():
    try:
        result = run('avendesora help misdirection')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        One way to avoid being compelled to disclose a secret is to disavow
        any knowledge of the secret.  However, the presence of an account in
        Avendesora that pertains to that secret undercuts this argument.
        This is the purpose of stealth accounts. They allow you to generate
        secrets for accounts for which Avendesora has no stored information.
        In this case Avendesora asks you for the minimal amount of
        information that it needs to generate the secret. However in some
        cases, the amount of information that must be retained is simply too
        much to keep in your head. In that case another approach, referred
        to as secret misdirection, can be used.

        With secret misdirection, you do not disavow any knowledge of the
        secret, instead you say your knowledge is out of date. So you would
        say something like "I changed the password and then forgot it", or
        "The account is closed". To support this ruse, you must use the
        --seed (or -S) option to 'avendsora value' when generating your
        secret (secrets misdirection only works with generated passwords,
        not stored passwords). This causes Avendesora to ask you for an
        additional seed at the time you request the secret. If you do not
        use --seed or you do and give the wrong seed, you will get a
        different value for your secret.  In effect, using --seed when
        generating the original value of the secret causes Avendesora to
        generate the wrong secret by default, allowing you to say "See, I
        told you it would not work". But when you want it to work, you just
        interactively provide the correct seed.

        You would typically only use misdirection for secrets you are
        worried about being compelled to disclose. So it behooves you to use
        an unpredictable additional seed for these secrets to reduce the
        chance someone could guess it.

        Be aware that when you employ misdirection on a secret, the value of
        the secret stored in the archive will not be the true value, it
        will instead be the misdirected value.
    """).strip()
    assert result.decode('utf-8') == expected

# test_overview() {{{1
def test_overview():
    try:
        result = run('avendesora help overview')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Avendesora is password utility that can store your account
        information and generate account passwords and produce them from the
        command line.  It can also be configured to autotype your username
        and password into the current window so that you can log in with a
        simple keystroke.

        Avendesora is capable of generating passwords (character-based
        passcodes) or pass phrases (word-based passcodes).  Pass phrases are
        generally preferred if you have a choice, but many websites will not
        take them.  The benefit of pass phrases is that they are relatively
        easy to remember and type, and they are very secure.  The pass
        phrases generated by Avendesora generally consist of four words,
        each word is drawn from a dictionary of 10,000 words.  Thus, even if
        a bad guy knew that four lower case words were being used for your
        pass phrase, there are still 10,000,000,000,000,000 possible
        combinations for him to try (this represents a minimum entropy of 53
        bits).  Using six words results in 80 bits of entropy, which meets
        the threshold recommended by NIST for the most secure pass phrases.
        For more on this, see 'avendesora help entropy'.

        For another perspective on the attractiveness of pass phrases, see
        http://xkcd.com/936/.

        Unlike password vaults, Avendesora produces a highly unpredictable
        password from a master seed and the name of the account for which
        the password is to be used. The process is completely repeatable. If
        you give the same master seed and account name, you will get the
        same password. As such, the passwords do not have to be saved;
        instead they are regenerated on the fly.

        As a password generator, Avendesora provides three important
        advantages over conventional password vaults.  First, it allows
        groups of people to share access to accounts without having to
        securely share each password.  Instead, one member of the group
        creates a master seed that is securely shared with the group once.
        From then on any member of the group can create a new account, share
        the name of the account, and all members will know the password
        needed to access the account. The second advantage is that it opens
        up the possibility of using high-quality passwords for stealth
        accounts, which are accounts where you remember the name of the
        account but do not store any information about even the existence of
        the account on your computer.  With Avendesora, you only need to
        remember the name of the account and it will regenerate the password
        for you. This is perfect for your TrueCrypt hidden volume password.
        Finally, by securely storing a small amount of information, perhaps
        on a piece of paper in your safe-deposit box, you can often recover
        most if not all of your passwords even if you somehow lose your
        accounts file. You can even recover passwords that were created
        after you created your backup. This is because Avendesora combines
        the master seed with some easily reconstructed information, such as
        the account name, to create the password. If you save the master
        seed, the rest should be recoverable.

        To use Avendesora, one creates one or more accounts files. At a
        minimum, there is should be one for your private accounts, and one
        each for all the people with whom you hold shared accounts.  These
        files contain information about each of your non-stealth accounts.
        Avendesora then provides access to that information.  The
        information will include both secrets and non-secrets. The secrets
        can be given explicitly, or information could be provided to specify
        how to generate the secret.
    """).strip()
    assert result.decode('utf-8') == expected

# test_phishing() {{{1
def test_phishing():
    try:
        result = run('avendesora help phishing')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Phishing is a very common method used on the web to get people to
        unknowingly divulge sensitive information such as account
        credentials.  It is generally accomplished by sending misleading
        URLs in email or placing them on websites. When you visit these
        URLs you are taken to a site that looks identical to the site you
        were expecting to go to in the hope that you are tricked into giving
        up your account credentials.  It used to be that if you carefully
        inspected the url you could spot deception, but even that is no
        longer true.

        Avendesora helps you avoid phishing attacks in two ways. First, you
        should never go to one of your secure sites by clicking on a link.
        Instead, you should use Avendesora's browse command:

            avendesora browse chase

        In this way you use the URL stored in Avendesora rather than
        trusting a url link provided by a third party. Second, you should
        auto-type the account credentials using Avendesora's account
        discovery based on RecognizeURL() (be sure to use RecognizeURL() for
        websites rather than RecognizeTitle() when configuring account
        discovery). RecogniseURL() will not be fooled by a phishing site).
    """).strip()
    assert result.decode('utf-8') == expected

# test_questions() {{{1
def test_security_questions():
    try:
        result = run('avendesora help sec_quest')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Security questions are form of security theater imposed upon you by
        many websites. The claim is that these questions increase the
        security of your account. In fact they often do the opposite by
        creating additional avenues of access to your account. Their real
        purpose is to allow you to regain access to your account in case you
        lose your password. If you are careful, this is not needed (you do
        back up your Avendesora accounts, right?). In this case it is better
        to randomly generate your answers.

        Security questions are handled by adding something like the
        following to your account:

            questions = [
                Question('oldest aunt'),
                Question('title of first job'),
                Question('oldest uncle'),
                Question('savings goal'),
                Question('childhood vacation spot'),
            ]

        The string identifying the question does not need to contain the
        question verbatim, a abbreviated version is sufficient as long as it
        allows you to distinguish the question.  The questions are given as
        an array, and so are accessed with an index that starts at 0. Thus,
        to get the answer to who is your 'oldest aunt', you would use:

            > avendesora value <accountname> 0
            questions.0 (oldest aunt): ampere reimburse duster

        You can get a list of your questions so you can identify which index
        to use with:

            > avendesora values <accountname>
            ...
            questions:
                0: oldest aunt <reveal with 'avendesora value <accountname> questions.0'>
                1: title of first job <reveal with 'avendesora value <accountname> questions.1'>
                2: oldest uncle <reveal with 'avendesora value <accountname> questions.2'>
                3: savings goal <reveal with 'avendesora value <accountname> questions.3'>
                4: childhood vacation spot <reveal with 'avendesora value <accountname> questions.4'>
            ...

        By default, Avendesora generates a response that consists of 3
        random words. This makes it easy to read to a person over the phone
        if asked to confirm your identity.  Occasionally you will not be
        able to enter your own answer, but must choose one that is offered
        to you. In this case, you can specify the answer as part of the
        question:

            questions = [
                Question('favorite fruit', answer='grapes'),
                Question('first major city visited', answer='paris'),
                Question('favorite subject', answer='history'),
            ]

        When giving the answers you may want to conceal them to protect them
        from casual observation.

        Be aware that the question is used as a seed when generating the
        answer, so if you change the question in any way it changes the
        answer.
    """).strip()
    assert result.decode('utf-8') == expected

# test_scripts() {{{1
def test_scripts():
    try:
        result = run('avendesora help scripts')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Scripts are strings that contain embedded account attributes.  For
        example:

            'username: {username}, password: {passcode}'

        When processed by Advendesora the attributes are replaced by their
        value from the chosen account.  For example, this script might
        be rendered as:

            username: jman, password: R7ibHyPjWtG2

        Scripts are useful if you need to combine an account value with
        other text, if you need to combine more than one account value, or
        if you want quick access to something that would otherwise need an
        additional key.

        For example, consider an account for your wireless router, which
        might hold several passwords, one for administrative access and one
        or more for the network passwords.  Such an account might look like:

            class WiFi(Account):
                username = 'admin'
                passcode = Passphrase()
                networks = ["Occam's Router", "Occam's Router (guest)"]
                network_passwords = [Passphrase(), Passphrase()]
                privileged = Script('SSID: {networks.0}, password: {network_passwords.0}')
                guest = Script('SSID: {networks.1}, password: {network_passwords.1}')

        Now the credentials for the privileged network are accessed with:

            > avendesora value wifi privileged
            SSID: Occam's Router, password: overdraw cactus devotion saying

        Most account attributes that expect a string can also accept a
        script given in this manner.

        You can also give a script rather than a field on the command line
        when running the value command:

            > avendesora value scc '{username}: {passcode}'
            jman: R7ibHyPjWtG2

        It is also possible to specify a script for the value of the
        *default* attribute. This attribute allows you to specify the
        default field (which attribute name and key to use if one is not
        given on the command line).  It also accepts a script rather than a
        field, but in this case it should be a simple string and not an
        instance of the Script class.  If you passed it as a Script, it
        would be expanded before being interpreted as a field name, and so
        would result in a 'not found' error.

            class SCC(Acount):
                aliases = 'scc'
                username = 'jman'
                password = PasswordRecipe('12 2u 2d 2s')
                default = 'username: {username}, password: {password}'

        You can access the script by simply not providing a field.

            > avendesora value scc
            username: jman, password: *m7Aqj=XBAs7

        Finally, you pass a script to the account discovery recognizers.
        They specify the action that should be taken when a particular
        recognizer triggers. These scripts would also be simple strings and
        not instances of the Script class. For example, this recognizer
        could be used to recognize Gmail:

            discovery = [
                RecognizeURL(
                    'https://accounts.google.com/ServiceLogin',
                    script='{username}{return}{sleep 1.5}{passcode}{return}'
                ),
                RecognizeURL(
                    'https://accounts.google.com/signin/challenge',
                    script='{questions.0}{return}'
                ),
            ]

        Besides the account attributes, you can use several other special
        attributes including: {tab}, {return}, and {sleep N}.  {tab} is
        replaced by a tab character, {return} is replaced by a carriage
        return character, and {sleep N} causes a pause of N seconds. The
        sleep function is only active when autotyping after account
        discovery.
    """).strip()
    assert result.decode('utf-8') == expected

# test_secrets() {{{1
def test_secrets():
    try:
        result = run('avendesora help secrets')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Secrets can either by obscured or generated.

        Obscured Secrets
        ================

        Obscured secrets are secrets that are those that are given to Avendesora
        to securely hold. The may be things like account numbers or existing
        passwords.  There are several ways for Avendesora to hold a secret,
        presented in order of increasing security.

        Hide
        ----

        This marks a value as being confidential, meaning that it will be
        protected when shown to the user, but value is not encoded or encrypted
        in any way.  Rather, it accounts on the protections afforded the
        accounts file to protect its secret.

            Hide(plaintext, secure=True)

            plaintext (str):
                The secret in plain text.
            secure (bool):
                Indicates that this secret should only be contained in an
                encrypted accounts file. Default is True.

        Example:

            account = Hide('5206-7844')


        Hidden
        -------

        This obscures but does not encrypt the text. It can protect the secret from
        observers that get a quick glance of the encoded text, but if they are able to
        capture it they can easily decode it.

            Hidden(encodetext, secure=True, encoding=None)

            plaintext (str):
                The secret encoded in base64.
            secure (bool):
                Indicates that this secret should only be contained in an
                encrypted accounts file. Default is True.
            encoding (str):
                The encoding to use for the deciphered text.

        Example:

            account = Hidden('NTIwNi03ODQ0')

        To generate the encoded text, use:

            > avendesora conceal


        GPG
        ---

        The secret is fully encrypted with GPG. Both symmetric encryption and
        key-based encryption are supported.  This can be used to protect a
        secret held in an unencrypted account file, in which case encrypting
        with your key is generally preferred. It can also be used to further
        protect a extremely valuable secret, in which case symmetric encryption
        is generally used.

            GPG(ciphertext, encoding=None)

            ciphertext (str):
                The secret encrypted and armored by GPG.
            encoding (str):
                The encoding to use for the deciphered text.

        Example:

            secret = GPG('''
                -----BEGIN PGP MESSAGE-----
                Version: GnuPG v2.0.22 (GNU/Linux)

                jA0ECQMCwG/vVambFjfX0kkBMfXYyKvAuCbT3IrEuEKD//yuEMCikciteWjrFlYD
                ntosdZ4WcPrFrV2VzcIIcEtU7+t1Ay+bWotPX9pgBQcdnSBQwr34PuZi
                =4on3
                -----END PGP MESSAGE-----
            ''')

        To generate the cipher text, use:

            > avendesora conceal -e gpg

        The benefit of using symmetric GPG encryption on a secret that is
        contained in an encrypted account file is that the passphrase will
        generally not be found in the GPG agent, in which case someone could not
        walk up to your computer while your screen is unlocked and successfully
        request the secret.  However, the GPG agent does retain the password for
        a while after you decrypt the secret. If you are concerned about that,
        you should follow your use of Avendesora with the following command,
        which clears the GPG agent:

            > killall gpg-agent


        Scrypt
        ------

        The secret is fully encrypted with Scrypt. You personal Avendesora
        encryption key is used (contained in ~/.config/avendesora/.key.gpg). As
        such, these secrets cannot be shared. This encryption method is only
        available if you have installed scrypt on your system (pip3 install
        --user scrypt). Since the Scrypt class only exists if you have installed
        scrypt, it is not imported into your accounts file. You would need to
        import it yourself before using it.

            Script(ciphertext, encoding=None)

            ciphertext (str):
                The secret encrypted by scrypt.
            encoding (str):
                The encoding to use for the deciphered text.

        Example:
            from avendesora import Scrypt
            ...
            secret = Scrypt(c2NyeXB0ABAAAAAIAAAAASfBZvtYnHvgdts2jrz5RfbYlFYj/EQgiM1IYTnXKHhMkleZceDg0yUaOWa9PzmZueppNIzVdawAOd9eSVgGeZAIh4ulPHPBGAzXGyLKc/vo8Fe24JnLr/RQBlTjM9+r6vbhi6HFUHD11M6Ume8/0UGDkZ0=)

        To generate the cipher text, use:

            > avendesora conceal -e scrypt


        Generated Secrets
        =================

        Generated secrets are secrets for which the actual value is arbitrary,
        but it must be quite unpredictable. Generated secrets are generally used
        for passwords and pass phrases, but it can also be used for things like
        personal information requested by institutions that they have no need to
        know. For example, a website might request your birth date to assure
        that you are an adult, but then also use it as a piece of identifying
        information if you ever call and request support.  In this case they do
        not need your actual birth date, they just need you to give the same
        date every time you call in.


        Password
        --------

        Generates an arbitrary password by selecting symbols from the given
        alphabet at random. The entropy of the generated password is
        length*log2(len(alphabet)).

            Password(
                length=12, alphabet=DISTINGUISHABLE, master=None, version=None,
                sep='', prefix='', suffix=''
            )

            length (int):
                The number of items to draw from the alphabet when creating the
                password.  When using the default alphabet, this will be the
                number of characters in the password.
            alphabet (str):
                The reservoir of legal symbols to use when creating the
                password. By default the set of easily distinguished
                alphanumeric characters are used. Typically you would use the
                pre-imported character sets to construct the alphabet. For
                example, you might pass:
                    ALPHANUMERIC + '+=_&%#@'
            master (str):
                Overrides the master seed that is used when generating the
                password.  Generally, there is one master seed shared by all
                accounts contained in an account file.  This argument overrides
                that behavior and instead explicitly specifies the master seed
                for this secret.
            version (str):
                An optional seed. Changing this value will change the generated
                password.
            shift_sort(bool):
                If true, the characters in the password will be sorted so that
                the characters that require the shift key when typing are placed
                last, making it easier to type. Use this option if you expect to
                be typing the password by hand.
            sep (str):
                A string that is placed between each symbol in the generated
                password.
            prefix (str):
                A string added to the front of the generated password.
            suffix (str):
                A string added to the end of the generated password.

        Example:

            passcode = Password(10)


        Passphrase
        ----------

        Similar to Password in that it generates an arbitrary pass phrase by
        selecting symbols from the given alphabet at random, but in this case
        the default alphabet is a dictionary containing about 10,000 words.

            Passphrase(
                length=4, alphabet=None, master=None, version=None, sep=' ', prefix='',
                suffix=''
            )

            length (int):
                The number of items to draw from the alphabet when creating the
                password.  When using the default alphabet, this will be the
                number of words in the passphrase.
            alphabet (str):
                The reservoir of legal symbols to use when creating the
                password. By default, this is a predefined list of 10,000 words.
            master (str):
                Overrides the master seed that is used when generating the
                password.  Generally, there is one master seed shared by all
                accounts contained in an account file.  This argument overrides
                that behavior and instead explicitly specifies the master seed
                for this secret.
            version (str):
                An optional seed. Changing this value will change the generated
                pass phrase.
            sep (str):
                A string that is placed between each symbol in the generated
                password.
            prefix (str):
                A string added to the front of the generated password.
            suffix (str):
                A string added to the end of the generated password.

        Example:

            passcode = Passphrase()


        PIN
        ---

        Similar to Password in that it generates an arbitrary PIN by selecting
        symbols from the given alphabet at random, but in this case the default
        alphabet is the set of digits (0-9).

            PIN(length=4, alphabet=DIGITS, master=None, version=None)

            length (int):
                The number of items to draw from the alphabet when creating the
                password.  When using the default alphabet, this will be the
                number of digits in the PIN.
            alphabet (str):
                The reservoir of legal symbols to use when creating the
                password. By default the digits (0-9) are used.
            master (str):
                Overrides the master seed that is used when generating the
                password.  Generally, there is one master seed shared by all
                accounts contained in an account file.  This argument overrides
                that behavior and instead explicitly specifies the master seed
                for this secret.
            version (str):
                An optional seed. Changing this value will change the generated
                PIN.

        Example:

            passcode = PIN()


        Question
        --------

        Generates an arbitrary answer to a given question. Used for website
        security questions. When asked one of these security questions it can be
        better to use an arbitrary answer. Doing so protects you against people
        who know your past well and might be able to answer the questions.

        Similar to Passphrase() except a question must be specified when created
        and it is taken to be the security question. The question is used rather
        than the field name when generating the secret.

            Question(
                question, length=3, alphabet=None, master=None, version=None,
                sep=' ', prefix='', suffix='', answer=None
            )

            question (str):
                The question to be answered. Be careful. Changing the question
                in any way will change the resulting answer.
            length (int):
                The number of items to draw from the alphabet when creating the
                password. When using the default alphabet, this will be the
                number of words in the answer.
            alphabet (list of strs):
                The reservoir of legal symbols to use when creating the
                password. By default, this is a predefined list of 10,000 words.
            master (str):
                Overrides the master seed that is used when generating the
                password.  Generally, there is one master seed shared by all
                accounts contained in an account file.  This argument overrides
                that behavior and instead explicitly specifies the master seed
                for this secret.
            version (str):
                An optional seed. Changing this value will change the generated
                answer.
            sep (str):
                A string that is placed between each symbol in the generated
                password.
            prefix (str):
                A string added to the front of the generated password.
            suffix (str):
                A string added to the end of the generated password.
            answer:
                The answer. If provided, this would override the generated
                answer.  May be a string, or it may be an Obscured object.

        Example:

            questions = [
                Question('Favorite foreign city'),
                Question('Favorite breed of dog'),
            ]


        PasswordRecipe
        --------------

        Generates passwords that can conform to the restrictive requirements
        imposed by websites. Allows you to specify the length of your password,
        and how many characters should be of each type of character using a
        recipe. The recipe takes the form of a string that gives the total
        number of characters that should be generated, and then the number of
        characters that should be taken from particular character sets. The
        available character sets are:

        l - lower case letters (a-z)
        u - upper case letters (A-Z)
        d - digits (0-9)
        s - punctuation symbols
        c - explicitly given set of characters

        For example, '12 2u 2d 2s' is a recipe that would generate a
        12-character password where two characters would be chosen from the
        upper case letters, two would be digits, two would be punctuation
        symbols, and the rest would be alphanumeric characters. It might
        generate something like: *m7Aqj=XBAs7

        Using '12 2u 2d 2c!@#$%^&*' is similar, except the punctuation symbols
        are constrained to be taken from the given set that includes !@#$%^&*.
        It might generate something like: YO8K^68J9oC!

            PasswordRecipe(
                recipe, def_alphabet=ALPHANUMERIC, master=None, version=None,
            )

            recipe (str):
                A string that describes how the password should be constructed.
            def_alphabet (list of strs):
                The alphabet to use when filling up the password after all the
                constraints are satisfied.
            master (str):
                Overrides the master seed that is used when generating the
                password.  Generally, there is one master seed shared by all
                accounts contained in an account file.  This argument overrides
                that behavior and instead explicitly specifies the master seed
                for this secret.
            version (str):
                An optional seed. Changing this value will change the generated
                answer.
            shift_sort(bool):
                If true, the characters in the password will be sorted so that
                the characters that require the shift key when typing are placed
                last, making it easier to type. Use this option if you expect to
                be typing the password by hand.

        Example:

            passcode = PasswordRecipe('12 2u 2d 2c!@#$%^&*')


        BirthDate
        ---------

        Generates an arbitrary birthdate for someone in a specified age range.


            BirthDate(
                year, min_age=18, max_age=65, fmt='YYYY-MM-DD',
                master=None, version=None,
            )

            year (int):
                The year the age range was established.
            min_age (int):
                The lower bound of the age range.
            max_age (int):
                The upper bound of the age range.
            fmt (str):
                Specifies the way the date is formatted. Consider an example
                date of 6 July 1969. YY and YYYY are replaced by the year (69 or
                1969). M, MM, MMM, and MMMM are replaced by the month (7, 07,
                Jul, or July). D and DD are replaced by the day (6 or 06).
            master (str):
                Overrides the master seed that is used when generating the
                password.  Generally, there is one master seed shared by all
                accounts contained in an account file.  This argument overrides
                that behavior and instead explicitly specifies the master seed
                for this secret.
            version (str):
                An optional seed. Changing this value will change the generated
                answer.

        Example:

            birthdate = BirthDate(2015, 21, 55)


        OTP
        ---

        Generates a secret that changes once per minute that generally is used
        as a second factor when authenticating.  It can act as a replacement
        for, and is fully compatible with, Google Authenticator.  You would
        provide the text version of the shared secret (the backup code) that is
        presented to you when first configuring your second factor authentication.

        Example:

            otp = OTP('JBSWY3DPEHPK3PXP')
    """).strip()
    assert result.decode('utf-8') == expected

# test_stealth() {{{1
def test_stealth():
    try:
        result = run('avendesora help stealth')
    except OSError as err:
        result = os_error(err)
    expected = dedent("""
        Normally Avendesora uses information from an account that is
        contained in an account file to generate the secrets for that
        account. In some cases, the presence of the account itself, even
        though it is contained within an encrypted file can be problematic.
        The mere presence of an encrypted file may result in you being
        compelled to open it. For the most damaging secrets, it is best if
        there is no evidence that the secret exists at all. This is the
        purpose of stealth accounts. (Misdirection is an alternative to
        stealth accounts; see 'avendesora help misdirection').

        Generally one uses the predefined stealth accounts, which all have
        names that are descriptive of the form of the secret they generate,
        for example word4 generates a 4-word pass phrase (also referred as
        the xkcd pattern):

            > avendesora value word4
            account: my_secret_account
            gulch sleep scone halibut

        The predefined accounts are kept in
        ~/.config/avendesora/stealth_accounts.  You are free to add new
        accounts or modify the existing accounts.

        Stealth accounts are subclasses of the StealthAccount class. These
        accounts differ from normal accounts in that they do not contribute
        the account name to the secrets generators for use as a seed.
        Instead, the user is requested to provide the account name every
        time the secret is generated. The secret depends strongly
        on this account name, so it is essential you give precisely the same
        name each time. The term 'account name' is being use here, but you
        can enter any text you like.  Best to make this text very difficult
        to guess if you are concerned about being compelled to disclose your
        GPG keys.

        The secret generator will combine the account name with the master
        seed before generating the secret. This allows you to use simple
        predictable account names and still get an unpredictable secret.
        The master seed used is taken from master_seed in the file
        that contains the stealth account if it exists, or the user_key if
        it does not. By default the stealth accounts file does not contain a
        master seed, which makes it difficult to share stealth accounts.
        You can create additional stealth account files that do contain
        master seeds that you can share with your associates.
    """).strip()
    assert result.decode('utf-8') == expected

# test_urls() {{{1
def test_urls():
    # these tests are important because the documentation and the program may
    # have different ideas on what the URL for this help topics is.
    import requests
    from avendesora.command import Command
    from avendesora.help import HelpMessage
    base_url = 'https://avendesora.readthedocs.io/en/latest'

    def url_exists(url):
        anchor = url.split('#')[-1] if '#' in url else None

        # I have no idea why sphinx decides to use the target rather than the title 
        # for the anchor for the identity command, but here is a workaround.
        exceptions = {
            'identity-generate-an-identifying-response-to-a-challenge':
                'identity-command',
        }
        anchor = exceptions.get(anchor, anchor)

        try:
            r = requests.get(url)
            if r.status_code != 200:
                return False
            if anchor and 'href="#{anchor}"'.format(anchor=anchor) in r.text:
                return True
            return not anchor
        except:
            return False

    for cmd in Command.commands():
        path = cmd.get_help_url()
        url = base_url + path
        assert url_exists(url), url
    for topic in HelpMessage.topics():
        path = topic.URL
        url = base_url + path
        assert url_exists(url), url




# Account Files
#
# Loads account files either individually or en mass.
#
# Loading account files on demand makes Avendesora more responsive on most
# commands.
#
# Avendesora uses the manifests file, stored in configuration directory, to
# determine which file holds a particular account. This file contains a pickled
# manifests dictionary, which lists all account names and aliases contained in
# each account file. The account names and aliases are all hashed so it is not
# possible to know the names of the user's accounts by examining the manifests
# file. Once loaded the manifests mapping is inverted to create the account
# catalog, which maps account names hashes to file names.

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .account import Account, canonicalize
from .config import get_setting
from .error import PasswordError
from .gpg import PythonFile
from .shlib import getmod
from .utilities import OSErrors
from inform import comment, codicil, log, os_error, warn
from hashlib import md5
from textwrap import dedent, fill
from time import time
import pickle
import sys


# Globals {{{1
MANIFESTS_FILENAME = 'manifests'
if sys.version_info.major < 3:
    PICKLE_ARGS = dict()
else:
    PICKLE_ARGS = dict(encoding='utf8')



# AccountFiles class (private) {{{1
class AccountFiles:
    def __init__(self, warnings=True): # {{{2
        # initialize object {{{3
        self.loaded = set()
        self.catalog = {}
        self.shared_secrets = {}
        self.existing_names = {}

        # check permissions, dates on account files {{{3
        most_recently_updated = 0
        mask = get_setting('account_file_mask')
        for filename in get_setting('accounts_files', []):
            path = get_setting('settings_dir') / filename
            resolved_path = path.resolve()

            # check file permissions
            permissions = getmod(resolved_path)
            violation = permissions & mask
            if violation:
                recommended = permissions & ~mask & 0o777
                warn("file permissions are too loose.", culprit=path)
                codicil("Recommend running 'chmod {:o} {}'.".format(recommended, resolved_path))

            # determine time of most recently updated account file
            updated = resolved_path.stat().st_mtime
            if updated > most_recently_updated:
                most_recently_updated = updated

        # check for missing or stale archive file {{{3
        archive_file = get_setting('archive_file')
        if archive_file and warnings:
            if archive_file.exists():
                resolved_path = archive_file.resolve()

                # check file permissions
                permissions = getmod(resolved_path)
                violation = permissions & mask
                if violation:
                    recommended = permissions & ~mask & 0o777
                    warn("file permissions are too loose.", culprit=path)
                    codicil("Recommend running 'chmod {:o} {}'.".format(
                        recommended, resolved_path)
                    )

                # warn user if archive file is out of date
                stale = float(get_setting('archive_stale'))
                archive_updated = resolved_path.stat().st_mtime
                if most_recently_updated > archive_updated:
                    log('Avendesora archive is {:.0f} hours out of date.'.format(
                        (most_recently_updated - archive_updated)/3600
                    ))
                    age_in_seconds = time() - archive_updated
                    if age_in_seconds > 86400 * stale:
                        warn('stale archive.')
                        codicil(fill(dedent("""
                            Recommend running 'avendesora changed' to determine
                            which account entries have changed, and if all the
                            changes are expected, running 'avendesora archive'
                            to update the archive.
                        """).strip()))
                else:
                    log('Avendesora archive is up to date.')
            else:
                warn('archive missing.')
                codicil(
                    "Recommend running 'avendesora archive'",
                    "to create the archive."
                )

    def load_account_file(self, filename): # {{{2
        if filename in self.loaded:
            return

        # read the file
        settings_dir = get_setting('settings_dir')
        path = settings_dir / filename
        account_file = PythonFile(path)
        contents = account_file.run()
        master_seed = contents.get('master_seed')
        if master_seed:
            self.shared_secrets[path.stem] = master_seed

        # traverse through all accounts and pass in fileinfo and master seed;
        # will be ignored if already set
        for account in Account.all_accounts():
            account.preprocess(master_seed, account_file, self.existing_names)

        self.loaded.add(filename)

    def load_account(self, name): # {{{2
        canonical_name = canonicalize(name)
        self.read_manifests()
        hash = md5(canonical_name.encode('utf8')).digest()
        if hash in self.catalog:
            filename = self.catalog[hash]
            self.load_account_file(filename)
            assert canonical_name in Account._accounts
        else:
            # not in catalog, just read files until it is found
            for filename in get_setting('accounts_files', []):
                self.load_account_file(filename)
                if canonical_name in Account._accounts:
                    return
        self.write_manifests()

    def load_account_files(self): # {{{2
        for filename in get_setting('accounts_files', []):
            self.load_account_file(filename)
        self.write_manifests()

    def read_manifests(self): # {{{2
        if self.catalog:
            return
        settings_dir = get_setting('settings_dir')
        manifests_path = settings_dir / MANIFESTS_FILENAME
        try:
            contents = manifests_path.read_bytes()
            try:

                manifests = pickle.loads(contents, **PICKLE_ARGS)
                # build the catalog by inverting the manifests
                self.catalog = {
                    h:n for n,l in manifests.items() for h in l
                }
            except (ValueError, pickle.UnpicklingError) as e:
                warn('garbled manifest.', culprit=manifests_path, codicil=str(e))
            assert isinstance(self.catalog, dict)
        except OSErrors as e:
            comment(os_error(e))

    def write_manifests(self): # {{{2
        # do not modify existing catalog if no account files were loaded
        if not self.loaded:
            return
        log('Updating manifests.')

        # remove entries for loaded files from catalog
        if self.catalog:
            for k in self.catalog.keys():
                if k in self.loaded:
                    del self.catalog[k]

        # now add back the updated entries
        for name, account in Account._accounts.items():
            hash = md5(name.encode('utf8')).digest()
            self.catalog[hash] = account._file_info.path.name

        # build manifests by inverting the catalog
        manifests = {n:[] for n in get_setting('accounts_files', [])}
        for k, v in self.catalog.items():
            manifests[v].append(k)

        # write the manifests
        settings_dir = get_setting('settings_dir')
        manifests_path = settings_dir / MANIFESTS_FILENAME
        contents = pickle.dumps(manifests)

        try:
            manifests_path.write_bytes(contents)
        except OSError as e:
            warn(os_error(e))



# Account
# API for an account

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .browsers import StandardBrowser
from .collection import Collection
from .config import get_setting
from .error import PasswordError
from .obscure import ObscuredSecret
from .preferences import TOOL_FIELDS
from .recognize import Recognizer
from .script import Script
from .secrets import GeneratedSecret
from difflib import get_close_matches
from inform import (
    Color, codicil, conjoin, cull, full_stop, is_collection, is_str,
    join, log, notify, output, warn, indent,
)
from textwrap import dedent
try:
    from urllib.parse import urlparse
except ImportError:  # pragma: no cover
    from urlparse import urlparse
import re


# Globals {{{1
VECTOR_PATTERN = re.compile(r'\A(\w+)\[(\w+)\]\Z')
LabelColor = Color(
    color=get_setting('label_color'),
    scheme=get_setting('color_scheme'),
    enable=Color.isTTY()
)
HighlightColor = Color(
    color=get_setting('highlight_color'),
    scheme=get_setting('color_scheme'),
    enable=Color.isTTY()
)


# Utilities {{{1
def canonicalize(name):
    return name.replace('-', '_').lower()


# AccountValue class {{{1
class AccountValue:
    """An account value.

    This is the object returned by :meth:`avendesora.Account.get_value` and
    :meth:`avendesora.Account.get_values`.
    It contains information about a single account value. Specifically, it
    provides the following attributes: *value*, *is_secret*, *name*, *key*,
    *field*, and *desc*.
    """
    def __init__(self, value, is_secret, name=None, key=None, desc=None):
        self.value = value
        self.is_secret = is_secret
        self.name = name
        self.key = str(key) if key is not None else key
        self.field = '.'.join(cull([name, self.key]))
        self.desc = desc

    def __str__(self):
        "Returns value as string."
        if hasattr(self.value, 'entropy'):
            entropy = round(self.value.entropy)
            log('Entropy of {} = {} bits.'.format(self.name, entropy))
        return str(self.value)

    def render(self, fmts=('{f} ({d}): {v}', '{f}: {v}')):
        """Return value formatted as a string.

        Args:
            fmts (collection of strings):
                *fmts* contains a sequence of format strings that are tried in
                sequence.  The first one for which all keys are known is used.
                The possible keys are:

                | n -- name (identifier for the first level of a field)
                | k -- key (identifier for the second level of a field)
                | f -- field (name.key)
                | d -- description
                | v -- value

                If none work, the value alone is returned.
        Returns:
            The value rendered as a string.
        """
        value = str(self.value)
        if '\n' in value:
            value = '\n'+indent(dedent(value), get_setting('indent')).strip('\n')
        if is_str(fmts):
            fmts = fmts,

        # build list of arguments, deleting any that is not set
        args = {
            k: v for k, v in [
                ('f', self.field),
                ('k', self.key),
                ('n', self.name),
                ('d', self.desc),
                ('v', value)
            ] if v
        }

        # format the arguments, use first format sting that works
        for fmt in fmts:
            try:
                return fmt.format(**args)
            except KeyError:
                pass

        # nothing worked, just return the value
        return value

    def __iter__(self):
        "Cast AccountValue to a tuple to get value, is_secret, name, and desc."
        for each in [str(self), self.is_secret, self.name, self.desc]:
            yield each


# Account class {{{1
class Account(object):
    """Class that holds all the information specific to an account.

    Add desired account information as attributes of the class.
    """

    __NO_MASTER = True
        # prevents master seed from being added to this base class
    _accounts = {}

    # all_accounts() {{{2
    @classmethod
    def all_accounts(cls):
        for sub in cls.__subclasses__():
            # do not yield the base StealthAccount
            if sub != StealthAccount:
                yield sub
            for each in sub.all_accounts():
                yield each

    # get_account() {{{2
    @staticmethod
    def get_account(name):
        canonical = canonicalize(name)
        try:
            return Account._accounts[canonical]
        except KeyError:
            # account not found, assemble message giving suggested account names
            names = Account._accounts.keys()
            candidates = get_close_matches(canonical, names, 9, 0.6)

            # do not want to give multiple options all of which are aliases for
            # the same accounts, so look for up to three unique accounts
            close_matches = []
            for candidate in candidates:
                account = Account._accounts[candidate]
                if account not in close_matches:
                    close_matches.append(account)
                if len(close_matches) >= 3:
                    break

            # add the aliases to be base account names if available
            offer = []
            for account in close_matches:
                aliases = getattr(account, 'aliases', None)
                if aliases:
                    offer.append('{} ({})'.format(
                        account.get_name(), ', '.join(aliases)
                    ))
                else:
                    offer.append(account.get_name())

            # generate the message handling 0, 1, 2 or 3 candidates gracefully
            msg = 'account not found'
            if offer:
                msg = '{}, did you mean:\n    {}?'.format(
                    msg, conjoin(offer, sep=',\n    ', conj=', or\n    ')
                )
            raise PasswordError(full_stop(msg), culprit=name)

    # get_name() {{{2
    @classmethod
    def get_name(cls):
        """Get account name.

        Returns:
            Returns the primary account name. This is generally the class name
            converted to lower case unless it was overridden with the NAME
            attribute.
        """

        try:
            return cls.NAME
        except AttributeError:
            # consider converting lower to upper case transitions in __name__ to
            # dashes.
            return cls.__name__.lower()

    # get_seed() {{{2
    @classmethod
    def get_seed(cls):
        try:
            return cls.account_seed
        except AttributeError:
            return cls.get_name()

    # override_master() {{{2
    @classmethod
    def request_seed(cls):
        return getattr(cls, '_interactive_seed', False)

    # preprocess() {{{2
    @classmethod
    def preprocess(cls, master, fileinfo, seen):

        # return if this account has already been processed
        if hasattr(cls, '_file_info'):
            return  # don't override on those accounts where it was already set

        # add fileinfo
        cls._file_info = fileinfo

        # add master seed
        if master and not hasattr(cls, '_%s__NO_MASTER' % cls.__name__):
            if not hasattr(cls, 'master_seed'):
                cls.master_seed = master
                cls._master_source = 'file'
            else:
                cls._master_source = 'account'

        # convert aliases to a list
        if hasattr(cls, 'aliases'):
            aliases = list(Collection(cls.aliases))
            cls.aliases = aliases
        else:
            aliases = []

        # canonicalize names and look for duplicates
        new = {}
        account_name = cls.get_name()
        path = cls._file_info.path
        for name in [account_name] + aliases:
            canonical = canonicalize(name)
            Account._accounts[canonical] = cls
            if canonical in seen:
                if name == account_name:
                    warn('duplicate account name.', culprit=name)
                else:
                    warn('alias duplicates existing name.', culprit=name)
                codicil('Seen in %s in %s.' % seen[canonical])
                codicil('And in %s in %s.' % (account_name, path))
                break
            else:
                new[canonical] = (account_name, path)
        seen.update(new)
            # this two step approach to updating seen prevents us from
            # complaining about aliases that duplicate the account name,
            # or duplicate aliases, both of which are harmless

    # id_contains() {{{2
    @classmethod
    def id_contains(cls, target):
        target = target.lower()
        if target in cls.get_name().lower():
            return True
        try:
            for alias in cls.aliases:
                if target in alias.lower():
                    return True
        except AttributeError:
            pass
        return False

    # account_contains() {{{2
    @classmethod
    def account_contains(cls, target):
        if cls.id_contains(target):
            return True
        target = target.lower()
        for key, value in cls.items():
            try:
                if is_collection(value):
                    for k, v in Collection(value).items():
                        if target in v.lower():
                            return True
                elif target in value.lower():
                    return True
            except AttributeError:
                pass
        return False

    # recognize() {{{2
    @classmethod
    def recognize(cls, data, verbose):
        # try the specified recognizers
        discovery = getattr(cls, 'discovery', ())
        for recognizer in Collection(discovery):
            if isinstance(recognizer, Recognizer):
                script = recognizer.match(data, cls, verbose)
                name = getattr(recognizer, 'name', None)
                if script:
                    yield name, script
        if discovery:
            return

        # If no recognizers specified, just check the urls
        for url in Collection(cls.get_scalar('urls', default=[])):
            url = url if '//' in url else ('//'+url)
            url_components = urlparse(url)
            if url_components.scheme:
                protocol = url_components.scheme
            else:
                protocol = get_setting('default_protocol')
            host = url_components.netloc
            if host == data.get('host'):
                if (protocol == data.get('protocol')):
                    if verbose:
                        log('    %s: matches.' % cls.get_name())
                    yield None, True
                    return
                else:
                    msg = 'URL matches, but uses wrong protocol.'
                    notify(msg)
                    raise PasswordError(msg, culprit=cls.get_name())

    # initialize() {{{2
    @classmethod
    def initialize(cls, interactive_seed=False, stealth_name=None):
        cls._interactive_seed = interactive_seed
        log('Initializing account:', cls.get_name())
        try:
            if cls.master_seed.is_secure():
                if not cls._file_info.encrypted:
                    warn(
                        'high value master seed not contained in encrypted',
                        'account file.', culprit=cls.get_name()
                    )
        except AttributeError:
            pass
        for bad, good in get_setting('commonly_mistaken_attributes').items():
            if hasattr(cls, bad):
                warn(
                    '{} attribute found,'.format(bad),
                    'should probably be {}.'.format(good),
                    culprit=cls.get_name()
                )

        # do some more error checking
        if isinstance(getattr(cls, 'master_seed', None), GeneratedSecret):
            raise PasswordError(
                'master must not be a subclass of GeneratedSecret.',
                culprit=cls.get_name()
            )

    # keys() {{{2
    @classmethod
    def fields(cls, all=False):
        for field in sorted(cls.__dict__):
            if not field.startswith('_') and (all or field not in TOOL_FIELDS):
                yield field

    # items() {{{2
    @classmethod
    def items(cls, all=False):
        for field in cls.fields(all):
            yield field, getattr(cls, field)

    # get_fields() {{{2
    @classmethod
    def get_fields(cls, all=False):
        """Iterate through fields.

        Iterates through the field names.

        Example::

            for name, keys in account.get_fields():
                if keys:
                    display(name + ':')
                    for key, value in account.get_values(name):
                        display(indent(
                            value.render(('{k}) {d}: {v}', '{k}: {v}'))
                        ))
                else:
                    value = account.get_value(name)
                    display(value.render('{n}: {v}'))

        Args:
            all (bool):
                If False, ignore the tool fields.

        Returns:
            A pair (2-tuple) that contains both field name and the key names.
            None is returned for the key names if the field holds a scalar value.
        """
        for field in sorted(cls.__dict__):
            if not field.startswith('_') and (all or field not in TOOL_FIELDS):
                value = getattr(cls, field)
                if is_collection(value):
                    yield field, Collection(value).keys()
                else:
                    yield field, None

    # get_scalar() {{{2
    @classmethod
    def get_scalar(cls, name, key=None, default=False):
        """Get field Value given a field name and key.

        A lower level interface than *get_value()* that given a name and perhaps
        a key returns a scalar value.  Also takes an optional default value that is
        returned if the value is not found. Unlike *get_value()*, the actual value
        is returned, not a object that contains multiple facets of the value.

        The *name* is the field name, and the *key* would identity which value is
        desired if the field is a composite. If default is False, an error is raised
        if the value is not present, otherwise the default value itself is returned.

        Args:
            name (str):
                The name of the field.
            key (str or int)
                The key for the desired value (should be None for scalar values).
            default:
                The value to return if the requested value is not available.

        Returns:
            The requested value.
        """
        value = getattr(cls, name, None)
        if value is None:
            if name == 'NAME':
                return cls.get_name()
            if default is False:
                raise PasswordError(
                    'field not found.',
                    culprit=(cls.get_name(), cls.combine_field(name, key))
                )
            return default

        if key is None:
            if is_collection(value):
                choices = {}
                for k, v in Collection(value).items():
                    try:
                        desc = v.get_description()
                    except AttributeError:
                        desc = None
                    if desc:
                        choices['   %s: %s' % (k, desc)] = k
                    else:
                        choices['   %s:' % k] = k
                raise PasswordError(
                    'composite value found, need key. Choose from:',
                    *sorted(choices.keys()),
                    sep = '\n',
                    culprit=(cls.get_name(), cls.combine_field(name, key)),
                    is_collection = True,
                    collection = choices
                )
        else:
            try:
                if is_collection(value):
                    value = value[key]
                else:
                    warn('not a composite value, key ignored.', culprit=name)
                    key = None
            except (IndexError, KeyError, TypeError):
                raise PasswordError(
                    'key not found.',
                    culprit=(cls.get_name(), cls.combine_field(name, key))
                )

        # initialize the value if needed
        try:
            value.initialize(cls, name, key)
                # if Secret or Script, initialize otherwise raise exception
        except AttributeError:
            pass
        return value

    # is_secret() {{{2
    @classmethod
    def is_secret(cls, name, key=None):
        value = cls.__dict__.get(name)
        if isinstance(value, Script):
            str(value)  # side effect of convert to string is computing is_secret
            return value.is_secret
        if key is None:
            try:
                return value.is_secret
            except AttributeError:
                return isinstance(value, (GeneratedSecret, ObscuredSecret))
        else:
            try:
                try:
                    return value[key].is_secret
                except AttributeError:
                    return isinstance(value[key], (GeneratedSecret, ObscuredSecret))
            except (IndexError, KeyError, TypeError):
                raise PasswordError(
                    'not found.', culprit=cls.combine_field(name, key)
                )

    # split_field() {{{2
    @classmethod
    def split_field(cls, field):
        # Account fields can either be scalars or composites (vectors or
        # dictionaries). This function takes a string or tuple (field) that the
        # user provides to specify which account value they wish and splits it
        # into a name and a key.  If the field is a scalar, the key will be
        # None.  Users request a value using one of the following forms:
        #     True: use default name
        #     field: scalar value (key=None)
        #     index: questions (field->'questions', key=index)
        #     field[index] or field.index: for vector value
        #     field[key] or field.key: for dictionary value

        # handle empty field
        if field is True or field is False or field == '':
            field = None
        if field is None:
            field = cls.get_scalar('default', default=None)

        # convert field into integer if possible
        try:
            field = int(field)
        except (ValueError, TypeError):
            pass

        # separate field into name and key
        if type(field) is tuple:
            # split field if given as a tuple
            if len(field) == 1:
                name, key = field[0], None
            elif len(field) == 2:
                name, key = field[0], field[1]
            else:
                raise PasswordError('too many values.', culprit=field)
        elif is_str(field):
            # split field if given in the form: name[key]
            match = VECTOR_PATTERN.match(field)
            if match:
                name, key = match.groups()
            else:
                # split field if given in the form: name.key
                try:
                    name, key = field.split('.')
                except ValueError:
                    # must be simple name
                    name, key = field, None
        elif field is None:
            # handle defaulting
            defaults = get_setting('default_field').split()
            for default in defaults:
                if hasattr(cls, default):
                    name, key = default, None
                    break
            else:
                raise PasswordError(
                    'no default available, you must request a specific value.',
                    culprit=cls.get_name()
                )
        elif type(field) is int:
            name, key = get_setting('default_vector_field'), int(field)
        else:
            raise PasswordError('invalid field.', culprit=field)

        # look up name and key
        return cls.find_field(name, key)

    # find_field() {{{2
    @classmethod
    def find_field(cls, name, key=None):
        # look up field name while ignoring case and treating - and _ as same
        names = {
            n.replace('-', '_').lower(): n
            for n in cls.__dict__.keys()
            if not n.startswith('_')
        }
        try:
            name = names[name.replace('-', '_').lower()]
        except KeyError:
            raise PasswordError('field not found.', culprit=name)

        # name is now the true field name, now resolve key
        if key is None:
            return name, None
        try:
            return name, int(key)
        except ValueError:
            pass
        try:
            value = getattr(cls, name)
            keys = {
                n.replace('-', '_').lower(): n
                for n in value.keys()
            }
            try:
                key = keys[key.replace('-', '_').lower()]
            except KeyError:
                raise PasswordError('key not found.', culprit=key)
        except AttributeError:
            key = None
        return name, key

    # combine_field() {{{2
    @classmethod
    def combine_field(cls, name, key=None):
        # Inverse of split_field().

        if key is None:
            return name
        else:
            return '%s.%s' % (name, key)

    # get_value() {{{2
    @classmethod
    def get_value(cls, field=None):
        """Get account value.

        Return value from the account given a user friendly identifier or
        script. User friendly identifiers include:

         |  *None*: value of default field
         |  *name*: scalar value
         |  *name.key* or *name[key]*:
         |      member of a dictionary or array
         |      key is string for dictionary, integer for array

        Scripts are simply strings with embedded attributes.
        Ex: *'username: {username}, password: {passcode}'*

        Args:
            field (str):
                Field identifier or script.

        Returns:
            :class:`avendesora.AccountValue`: the desired value.
        """
        # get default if field was not given
        if not field:
            field = cls.get_scalar('default', default=None)

        # determine whether field is actually a script
        is_script = is_str(field) and '{' in field and '}' in field

        if is_script:
            # run the script
            script = Script(field)
            script.initialize(cls)
            value = str(script)
            is_secret = script.is_secret
            name = key = desc = None
        else:
            name, key = cls.split_field(field)

            value = cls.get_scalar(name, key)
            is_secret = cls.is_secret(name, key)
            try:
                desc = value.get_description()
            except AttributeError:
                desc = None
        return AccountValue(value, is_secret, name, key, desc)

    # get_values() {{{2
    @classmethod
    def get_values(cls, name):
        """Iterate through the values for a field.

        Args:
            name (str):
                The name of the field.

        Returns:
            Returns a pair (2-tuple) that contains the key and the value given
            as an :class:`avendesora.AccountValue` for each of the values.  If
            the value is a scalar, the key is None.
        """
        name, key = cls.find_field(name)
        value = getattr(cls, name, None)
        if value is None:
            if name == 'NAME':
                value = cls.get_name()
            else:
                return

        values = Collection(value, splitter=False)
        for key, val in values.items():
            value = cls.get_scalar(name, key)
            is_secret = cls.is_secret(name, key)
            try:
                desc = value.get_description()
            except AttributeError:
                desc = None
            yield key, AccountValue(value, is_secret, name, key, desc)

    # get_composite() {{{2
    @classmethod
    def get_composite(cls, name):
        """Get field value given a field name.

        A lower level interface than *get_value()* that given a name returns the
        value of the associated field, which may be a scalar (string or integer)
        or a composite (array of dictionary).  Unlike *get_value()*, the actual
        value is returned, not a object that contains multiple facets of the
        value.

        Args:
            name (str):
                The name of the field.

        Returns:
            The requested value.
        """
        value = getattr(cls, name, None)
        if value is None:
            if name == 'NAME':
                return cls.get_name()
            return None

        if is_collection(value):
            if hasattr(value, 'items'): # a dictionary or dictionary-like object
                result = {}
                for key in value.keys():
                    v = cls.get_scalar(name, key)
                    if isinstance(v, GeneratedSecret) or isinstance(v, ObscuredSecret):
                        v = str(v)
                    result[key] = v
            elif hasattr(value, 'append'):  # a list or list-like object
                result = []
                for index in range(len(value)):
                    v = cls.get_scalar(name, index)
                    if isinstance(v, GeneratedSecret) or isinstance(v, ObscuredSecret):
                        v = str(v)
                    result.append(v)
            else:
                raise NotImplementedError
        else:
            result = cls.get_scalar(name)
            if isinstance(result, GeneratedSecret) or isinstance(result, ObscuredSecret):
                result = str(result)
        return result

    # write_summary() {{{2
    @classmethod
    def write_summary(cls):
        # present all account values that are not explicitly secret to the user

        def fmt_field(name, value='', key=None, level=0):
            hl = False

            # resolve values
            if isinstance(value, Script):
                hl = True
                value = value.script
            elif cls.is_secret(name, key):
                reveal = "reveal with: {}".format(HighlightColor(join(
                    'avendesora',
                    'value',
                    cls.get_name(),
                    cls.combine_field(name, key)
                )))
                value = ', '.join(cull([value.get_description(), reveal]))
            elif isinstance(value, (GeneratedSecret, ObscuredSecret)):
                v = cls.get_scalar(name, key)
                value = ', '.join(cull([value.get_description(), str(v)]))
            else:
                value = str(value)

            # format values
            if '\n' in value:
                value = indent(dedent(value), get_setting('indent')).strip('\n')
                sep = '\n'
            elif value:
                sep = ' '
            else:
                sep = ''
            if hl:
                value = HighlightColor(value)
            name = str(name).replace('_', ' ')
            leader = level*get_setting('indent')
            return indent(
                LabelColor((name if key is None else str(key)) + ':') + sep + value,
                leader
            )

        # preload list with the names associated with this account
        names = [cls.get_name()] + getattr(cls, 'aliases', [])
        lines = [fmt_field('names', ', '.join(names))]

        for key, value in cls.items():
            if is_collection(value):
                lines.append(fmt_field(key))
                for k, v in Collection(value).items():
                    lines.append(fmt_field(key, v, k, level=1))
            else:
                lines.append(fmt_field(key, value))

        output(*lines, sep='\n')

    # archive() {{{2
    @classmethod
    def archive(cls):
        # return all account fields along with their values as a dictionary

        def extract(value, name, key=None):
            if not is_collection(value):
                if hasattr(value, 'initialize'):
                    value.initialize(cls, name, key)
                return value
            try:
                return {k: extract(v, name, k) for k, v in value.items()}
            except AttributeError:
                return [extract(v, name, i) for i, v in enumerate(value)]

        return {
            k: extract(v, k)
            for k, v in cls.items(True)
            if k not in ['master_seed', 'account_seed']
        }

    # open_browser() {{{2
    @classmethod
    def open_browser(cls, key=None, browser_name=None, list_urls=False):
        if not browser_name:
            browser_name = cls.get_scalar('browser', default=None)
        browser = StandardBrowser(browser_name)

        # get the urls from the urls attribute
        # this must be second so it overrides those from recognizers.
        primary_urls = getattr(cls, 'urls', [])
        if type(primary_urls) != dict:
            if is_str(primary_urls):
                primary_urls = primary_urls.split()
            primary_urls = {None: primary_urls} if primary_urls else {}

        # get the urls from the url recognizers
        discovery = getattr(cls, 'discovery', ())
        urls = {}
        for each in Collection(discovery):
            urls.update(each.all_urls())

        # combine, primary_urls must be added to urls, so they dominate
        urls.update(primary_urls)

        if list_urls:
            default = getattr(cls, 'default_url', None)
            for name, url in urls.items():
                if is_collection(url):
                    url = list(Collection(url))[0]
                if name == default:
                    url += HighlightColor(' [default]')
                    if not name:
                        name = ''
                elif not name:
                    continue
                output(LabelColor('{:>24s}:'.format(name)), url)
            return

        # select the urls
        keys = cull(list(urls.keys()))
        if not key:
            key = getattr(cls, 'default_url', None)
        if not key and keys and len(keys) == 1:
            key = keys[0]
        try:
            urls = urls[key]
        except KeyError:
            if keys:
                if key:
                    msg = 'unknown key, choose from {}.'
                else:
                    msg = 'key required, choose from {}.'
                raise PasswordError(
                    msg.format(conjoin(repr(k) for k in keys)), culprit=key
                )
            else:
                if key:
                    raise PasswordError(
                        'keys are not supported with urls on this account.',
                        culprit=key
                    )
                else:
                    raise PasswordError('no url available.')

        # open the url
        urls = Collection(urls)
        url = list(urls)[0]  # use the first url specified
        browser.run(url)

    # has_field() {{{2
    @classmethod
    def has_field(cls, name):
        return name in dir(cls)

    # get_username() {{{2
    @classmethod
    def get_username(cls):
        """Get the username.

        Like *get_value()*, but tries the *credential_ids* in order and returns
        the first found. *credential_ids* is an Avendesora configuration setting
        that by default is *username* and *email*.

        Returns:
            The username or email address.
        """
        identities = Collection(get_setting('credential_ids'))
        for identity in identities:
            try:
                return cls.get_value(identity)
            except PasswordError:
                pass

    # get_passcode() {{{2
    @classmethod
    def get_passcode(cls):
        """Get the passcode.

        Like *get_value()*, but tries the *credential_secrets* in order and returns
        the first found. *credential_secrets* is an Avendesora configuration setting
        that by default is *password*, *passphrase*, and *passcode*.

        Returns:
            The passcode.
        """
        secrets = Collection(get_setting('credential_secrets'))
        for passcode in secrets:
            try:
                return cls.get_value(passcode)
            except PasswordError:
                pass


# StealthAccount class {{{1
class StealthAccount(Account):
    """Empty account that defines how to generate a secret.

    With stealth accounts the actual account name is requested directly from the
    user. Only one attribute is generally given for stealth accounts, which is
    suitable as a default such as password, passphrase or passcode, and that
    contains the method for generating the desired secret based on the account
    name and the master seed in the stealth_accounts file.
    """

    __NO_MASTER = True
        # prevents master password from being added to this base class

    # initialize() {{{2
    @classmethod
    def initialize(cls, interactive_seed=False, stealth_name=None):
        cls._interactive_seed = interactive_seed
        cls._stealth_name = stealth_name
        log('Initializing stealth account:', cls.get_name())
        for key, value in cls.items():
            # reset the secrets so they honor stealth_name
            try:
                value.reset()
            except AttributeError:
                pass

    @classmethod
    def get_seed(cls):
        if cls._stealth_name:
            # In this case we are running using API rather than running from
            # command line and the account names was specified to get_account().
            return cls._stealth_name
        # need to handle case where stdin/stdout is not available.
        # perhaps write generic password getter that supports both gui and tui.
        # Then have global option that indicates which should be used.
        # Separate name from seed. Only request seed when generating a password.
        import getpass
        try:
            name = getpass.getpass('account name: ')
        except EOFError:
            output()
            name = ''
        if not name:
            warn('null account name.')
        return name

    @classmethod
    def archive(cls):
        # do not archive stealth accounts
        pass



"""Character Sets

Defines strings of characters and an exclude function that can be used either as
the alphabets for you character-base passwords or as building blocks used to
construct a new alphabet for you passwords.

Example:
To create an alphabet with all characters except tabs use either:
    'alphabet': exclude(PRINTABLE, '\t')
or:
    'alphabet': ALPHANUMERIC + PUNCTUATION + ' '
"""

# Exclude function
def exclude(chars, exclusions):
    """Exclude Characters

    Use this to remove characters from a character set.

    :arg str chars:
        Character set to strip.

    :arg str exclusions:
        Characters to remove from character set.

    Example::

        >>> exclude('ABCDEF', 'AEF')
        'BCD'

    """
    try:
        # this version is compatible with python3
        return chars.translate(str.maketrans('', '', exclusions))
    except AttributeError:
        # this version is compatible with python2
        return chars.translate(None, exclusions)

# Character sets
# Use these to construct alphabets by summing together the ones you want.

# Lower case ASCII letters.
LOWERCASE = "abcdefghijklmnopqrstuvwxyz"

# Upper case ASCII letters.
UPPERCASE = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

# Upper and lower case ASCII letters.
LETTERS = LOWERCASE + UPPERCASE

# ASCII digits.
DIGITS = "0123456789"

# ASCII letters and digits.
ALPHANUMERIC = LETTERS + DIGITS

# Hexidecimal digits.
HEXDIGITS = "0123456789abcdef"

# ASCII punctuation characters.
PUNCTUATION = r"""!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~"""

# ASCII punctuation characters excluding ', ", `, and \.
SYMBOLS = exclude(PUNCTUATION, """'"`\\""")

# ASCII white space characters (excluding newlines).
WHITESPACE = " \t"

# All ASCII printable characters (letters, digits, punctuation, whitespace).
PRINTABLE = ALPHANUMERIC + PUNCTUATION + WHITESPACE

# ASCII letters and digits with easily confused characters removed).
DISTINGUISHABLE = exclude(ALPHANUMERIC, 'Il1O0')

# ASCII characters that are typed using the shift key.
SHIFTED = UPPERCASE + '~!@#$%^&*()_+{}|:"<>?'



# shlib -- Scripting utilities
#
# A light-weight package with few dependencies that allows users to do 
# shell-script like things relatively easily in Python.

# License {{{1
# Copyright (C) 2016-2019 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see [http://www.gnu.org/licenses/].

__version__ = '1.0.0'
__released__ = '2019-01-13'

# Imports {{{1
try:
    from .extended_pathlib import Path
except ImportError:
    from pathlib import Path
import itertools
import shlex
import shutil
import errno
import os
import sys

# Parameters {{{1
PREFERENCES = dict(
    encoding = 'utf-8',
    log_cmd = False,
    use_inform = False,
)

# Utilities {{{1
# is_str {{{2
from six import string_types
def is_str(obj):
    """Identifies strings in all their various guises."""
    return isinstance(obj, string_types)

# is_iterable {{{2
import collections
def is_iterable(obj):
    """Identifies objects that can be iterated over, including strings."""
    return isinstance(obj, collections.Iterable)

# is_collection {{{2
def is_collection(obj):
    """Identifies objects that can be iterated over, excluding strings."""
    return is_iterable(obj) and not is_str(obj)

# to_path {{{2
def to_path(*args):
    try:
        return Path(*args).expanduser()
    except AttributeError:
        return Path(*args)

# to_paths {{{2
def to_paths(args):
    "Iterate through up to two levels of arguments, converting them to paths"
    for arg in args:
        if is_collection(arg):
            for each in arg:
                yield to_path(each)
        else:
            yield to_path(arg)
 
# to_str {{{2
def to_str(path):
    # first convert to path to assure ~ expansion is done, then convert back to 
    # string.
    return str(to_path(path))

# os_error {{{2
# Raise an error based on the errno.
def os_error(errno, filename=None):
    if filename:
        raise OSError(errno, os.strerror(errno), str(filename))
    else:
        raise OSError(errno, os.strerror(errno))

# split_cmd {{{2
def split_cmd(cmd):
    from shlex import split
    return split(cmd)

# quote_arg {{{2
def quote_arg(arg):
    """Return a shell-escaped version of the string *arg*."""
    try:
        from shlex import quote
    except ImportError:
        try:
            from pipes import quote
            import re
        except ImportError:
            def quote(arg):
                if not arg:
                    return "''"
                if re.search(r'[^\w@%+=:,./-]', arg, re.ASCII) is None:
                    return arg
                return "'" + arg.replace("'", "'\"'\"'") + "'"
    return quote(str(arg))

# _use_log {{{2
def _use_log(log):
    if log is None:
        return PREFERENCES['log_cmd']
    return log

# Preferences {{{1
def set_prefs(**kwargs):
    """Set ShLib preferences

    Args:
        use_inform (bool):
            Use inform for error reporting in the Cmd class and its subclasses.
            This provides a richer form of error reporting than simply using the
            OSError. Requires that inform be installed.
        log_cmd (bool):
            Log the command invocation and exit status in the Cmd class and its
            subclasses. Requires that inform be installed.
        encoding (str):
            The encoding to use if one is not specified.
    """
    PREFERENCES.update(kwargs)

# SHLib state
def get_state():
    return PREFERENCES

def set_state(state):
    global PREFERENCES
    old_state = PREFERENCES
    PREFERENCES = state
    return old_state

# File system utility functions (cp, mv, rm, ln, touch, mkdir, ls, etc.) {{{1
# cp {{{2
def cp(*paths):
    "Copy files or directories (equivalent to 'cp -rf')"
    dest = to_path(paths[-1])
    srcs = list(to_paths(paths[:-1]))
    if dest.is_dir():
        for src in srcs:
            if src.is_dir():
                fulldest = Path(dest, src.name)
                # required because dest cannot exist with copytree
                shutil.copytree(to_str(src), to_str(fulldest))
            else:
                shutil.copy2(to_str(src), to_str(dest))
        return
    if len(srcs) > 1:
        os_error(errno.ENOTDIR, dest)
    src = srcs[0]
    if src.is_dir() and dest.is_file():
        os_error(errno.EISDIR, src)
    if src.is_dir():
        # src is directory and dest does not exist
        shutil.copytree(to_str(src), to_str(dest))
    else:
        shutil.copy2(to_str(src), to_str(dest))

# mv {{{2
def mv(*paths):
    "Move file or directory (supports moves across filesystems)"
    dest = to_path(paths[-1])
    srcs = list(to_paths(paths[:-1]))
    assert len(srcs) >= 1
    if dest.is_dir():
        for src in srcs:
            fulldest = Path(dest, src.name)
            # required because dest cannot exist with shutil.move
            shutil.move(to_str(src), to_str(fulldest))
        return
    else:
        if len(srcs) > 1:
            os_error(errno.ENOTDIR, dest)
    src = srcs[0]
    if dest.is_file():
        if src.is_dir():
            os_error(errno.EISDIR, src)
        else:
            # overwrite destination
            shutil.move(to_str(src), to_str(dest))
    else:
        # destination does not exist
        assert not dest.exists()
        shutil.move(to_str(src), to_str(dest))


# rm {{{2
def rm(*paths):
    "Remove files or directories (equivalent to rm -rf)"
    for path in to_paths(paths):
        try:
            if path.is_dir():
                shutil.rmtree(to_str(path))
            else:
                path.unlink()
        # suitable for python3 only
        # except FileNotFoundError:
        #     pass
        except (IOError, OSError) as err:
            # don't complain if the file never existed
            if err.errno != errno.ENOENT:
                raise

# ln {{{2
def ln(src, dest):
    "Create symbolic link."
    dest = to_path(dest)
    dest.symlink_to(src)

# touch {{{2
def touch(*paths):
    """
    Touch one or more files. If files do not exist, create them.
    """
    for path in to_paths(paths):
        path.touch()

# mkdir {{{2
def mkdir(*paths):
    """
    Create a directory and all parent directories. Returns without complaint if
    directory already exists.
    """
    for path in to_paths(paths):
        try:
            path.mkdir(parents=True)
                # older versions of pathlib ignore exist_ok
                # in Python3.5, just add exist_ok=True to arg list
        # commented out version is suitable for Python3 only
        # except FileExistsError:
        #     if path.is_file():
        #         raise
        except (IOError, OSError) as err:
            if err.errno != errno.EEXIST or path.is_file():
                raise

# mount/umount {{{2
class mount:

    def __init__(self, path):
        self.path = to_path(path)
        self.mounted_externally = is_mounted(self.path)

        if not self.mounted_externally:
            Run(['mount', self.path])

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if not self.mounted_externally:
            umount(self.path)

def umount(path):
    Run(['umount', path])

def is_mounted(path):
    return Run(['mountpoint', '-q', path], '0,1').status == 0

# cd {{{2
class cd:
    def __init__(self, path):
        self.starting_dir = cwd()
        os.chdir(to_str(path))

    # support __enter__ and __exit__ so cd can be called in a with statement.
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, exc_traceback):
        os.chdir(to_str(self.starting_dir))

# cwd {{{2
def cwd():
    """Return current working directory as a pathlib path"""
    return Path.cwd()

# chmod {{{2
def chmod(mode, *paths):
    "Change the mode bits of one or more file or directory"
    for path in to_paths(paths):
        path.chmod(mode)

# getmod {{{2
def getmod(path):
    "Return the permission bits for a file or directory"
    return os.stat(str(path)).st_mode & 0o777

# ls {{{2
def ls(*paths, **kwargs):
    """
    List paths

    Paths may be files or directories. If path is a file, it is returned.
    If path is a directory, the items in the directory are returned.

    Args:
        paths: the paths to list ('.' if no paths given).
        select: a returned path will match this glob string, use **/* to enable 
            recursion
        reject: a returned path will not match this glob string
        only: specifies the type of returned paths, choose from 'file' or 'dir'
        hidden (bool): specifies whether hidden files should be returned, if 
            not given hidden files are returned if select string starts with 
            '.'

    KSK: it is a bit weird that I allow paths to be a list, but not select or
    reject. It would be nice to pass lists to both of those.

    Returns:
        path generator: iterates through filtered paths

    >>> from shlib import *

    >>> mkdir('d1', 'd2')
    >>> touch('d1/f1', 'd1/f2', 'd2/f1', 'd2/f2')

    >>> files = ls('d1')
    >>> set(str(f) for f in files) == set('d1/f1 d1/f2'.split())
    True

    >>> files = ls('d1', 'd2', select='*2')
    >>> set(str(f) for f in files) == set('d1/f2 d2/f2'.split())
    True

    >>> rm('d1', 'd2')

    """
    select = kwargs.get('select', '*')
    reject = kwargs.get('reject', '\0')
        # KSK: I have used '\0' as the default reject pattern because using ''
        # generates an exception. I put in an enhancement request to the pathlib
        # team to fix this, but it was rejected.
    only = kwargs.get('only')
    hidden = kwargs.get('hidden')

    def acceptable(path):
        if only == 'file' and not path.is_file():
            return False
        if only == 'dir' and not path.is_dir():
            return False
        if not retain_hidden and path.name.startswith('.'):
            return False
        if path.match(reject):
            return False
        return True

    select = to_str(select)
    retain_hidden = select.startswith('.') if hidden is None else hidden
    paths = paths if paths else ['.']
    for path in to_paths(paths):
        if path.is_file() and acceptable(path):
            if path.match(select):
                yield path
        elif path.is_dir():
            for each in path.glob(select):
                # glob() supports recursion so use it rather than iterdir()
                if acceptable(each):
                    yield each

# lsd {{{2
def lsd(*args, **kwargs):
    """
    List directories

    Same as ls with only='dir'.

    >>> dirs = lsd(select='xyz*')
    >>> set(str(d) for d in dirs) == set(''.split())
    True

    >>> mkdir('d1', 'd2')
    >>> touch('d1/f1', 'd1/f2', 'd2/f1', 'd2/f2')

    >>> dirs = lsd(select='d*')
    >>> set(str(d) for d in dirs) == set('d1 d2'.split())
    True

    >>> dirs = ls(select='*2')
    >>> set(str(d) for d in dirs) == set('d2'.split())
    True

    >>> rm('d1', 'd2')

    """
    kwargs['only'] = 'dir'
    for d in ls(*args, **kwargs):
        yield d

# lsf {{{2
def lsf(*args, **kwargs):
    """
    List files

    Same as ls with only='file'.

    >>> mkdir('d1', 'd2')
    >>> touch('d1/f1', 'd1/f2', 'd2/f1', 'd2/f2')

    >>> files = lsf('d1')
    >>> set(str(f) for f in files) == set('d1/f1 d1/f2'.split())
    True

    >>> files = lsf('d1', 'd2', select='f2')
    >>> set(str(f) for f in files) == set('d1/f2 d2/f2'.split())
    True

    >>> rm('d1', 'd2')

    """
    kwargs['only'] = 'file'
    for f in ls(*args, **kwargs):
        yield f

# Path list functions (leaves, cartesian_product, brace_expand, etc.) {{{1
def _leaves(path, hidden=False):
    for each in os.scandir(path):
        if each.is_dir(follow_symlinks=False):
            for e in _leaves(each):
                if hidden or not e.startswith('.'):
                    yield e
        else:
            if hidden or not each.name.startswith('.'):
                yield each.path

# leaves()  {{{2
def leaves(path, hidden=False):
    """Recursively descend into a directory yielding all of the file."""
    for each in _leaves(str(path)):
        yield Path(each)

# cartesian_product()  {{{2
def cartesian_product(*fragments):
    """
    Combine path fragments to to a path list. Each fragment must be a string or 
    path or an iterable that generates strings or paths.
    """
    if not len(fragments):
        return []
    return [Path(*f) for f in itertools.product(
        *(f if is_collection(f) else (f,) for f in fragments)
    )]

# brace_expand()  {{{2
try:
    from braceexpand import braceexpand

    def brace_expand(pattern):
        """Bash-style brace expansion"""
        for path in braceexpand(pattern):
            yield Path(path)

except ImportError:
    pass

# Execution classes and functions (Cmd, Run, Sh, Start, run, bg, shbg, which) {{{1
# Command class {{{2
class Cmd(object):
    # description {{{3
    """
    Specify a command

    cmd may be a string or a list.
    modes is a string that specifies various options
        S, s: Use, or do not use, shell
        O, o: Capture, or do not capture, stdout
        E, e: Capture, or do not capture, stderr
        M, m: Merge, or do not merge, stderr into stdout (M overrides E, e)
        W, s: Wait, or do not wait, for command to terminate before proceeding

        Only one of the following may be given, and it must be given last:
        *: accept any output status code
        N: accept any output status code equal to N or less
        M,N,...: accept status codes M, N, ...
    env is a dictionary of environment variable and their values.
    encoding is used on the input and output streams when converting them to and
        from strings.
    log specifies whether details about the command should be sent to log file.
        May be True, False, or None. If None, then behavior is set by log_cmd
        preference.
    option_args is used when rendering command to logfile, it indicates how many
        arguments each option takes.

    An exception is raised if exit status is not acceptable. By default an
    OSError is raised, however if the *use_inform* preference is true, then
    inform.Error is used. In this case the error includes attributes that can be
    used to access the stdout, stderr, status, cmd, and msg.
    """

    # __init__ {{{3
    def __init__(
        self, cmd, modes=None, env=None, encoding=None,
        log=None, option_args=None
    ):
        self.cmd = cmd
        self.env = env
        self.use_shell = False
        self.save_stdout = False
        self.save_stderr = False
        self.merge_stderr_into_stdout = False
        self.status = None
        self.wait_for_termination = True
        self.encoding = PREFERENCES['encoding'] if encoding is None else encoding
        self.log = log
        self.option_args = option_args
        self._interpret_modes(modes)

    # _interpret_modes {{{3
    def _interpret_modes(self, modes):
        accept = ''
        if modes:
            for i in range(len(modes)):
                mode = modes[i]
                if   mode == 's': self.use_shell = False
                elif mode == 'S': self.use_shell = True
                elif mode == 'o': self.save_stdout = False
                elif mode == 'O': self.save_stdout = True
                elif mode == 'e': self.save_stderr = False
                elif mode == 'E': self.save_stderr = True
                elif mode == 'm': self.merge_stderr_into_stdout = False
                elif mode == 'M': self.merge_stderr_into_stdout = True
                elif mode == 'w': self.wait_for_termination = False
                elif mode == 'W': self.wait_for_termination = True
                else: accept = modes[i:]; break
        self.accept = _Accept(accept)

    # run {{{3
    def run(self, stdin=None):
        """
        Run the command, will wait for it to terminate.

        If stdin is given, it should be a string. Otherwise, no connection is
        made to stdin of the command.

        Returns exit status if wait_for_termination is True.
        If wait_for_termination is False, you must call wait(), otherwise stdin
        is not be applied.  If you don't want to wait, call start() instead.
        """
        self.stdin = stdin
        import subprocess

        if is_str(self.cmd):
            cmd = self.cmd if self.use_shell else split_cmd(self.cmd)
        else:
            # cannot use to_str() because it can change some arguments when not intended.
            # this is particularly problematic the duplicity arguments in embalm
            cmd = [str(c) for c in self.cmd]
        if _use_log(self.log):
            from inform import log
            log('running:', render_command(cmd, option_args=self.option_args))

        # indicate streams to intercept
        streams = {}
        if stdin is not None:
            streams['stdin'] = subprocess.PIPE
        if self.save_stdout:
            streams['stdout'] = subprocess.PIPE
        if self.save_stderr:
            streams['stderr'] = subprocess.PIPE
        if self.merge_stderr_into_stdout:
            streams['stderr'] = subprocess.STDOUT

        # run the command
        try:
            process = subprocess.Popen(
                cmd, shell=self.use_shell, env=self.env, **streams
            )
        except OSError as e:
            if PREFERENCES['use_inform']:
                from inform import Error
                raise Error(
                    msg = os_error(e),
                    cmd = render_command(self.cmd),
                    template = '{msg}'
                )
            else:
                raise

        # store needed information and wait for termination if desired
        self.pid = process.pid
        self.process = process
        if self.wait_for_termination:
            return self.wait()

    # start {{{3
    def start(self, stdin=None):
        """
        Start the command, will not wait for it to terminate.

        If stdin is given, it should be a string. Otherwise, no connection is
        made to stdin of the command.
        """
        self.stdin = None
        import subprocess

        if is_str(self.cmd):
            cmd = self.cmd if self.use_shell else split_cmd(self.cmd)
        else:
            cmd = self.cmd
        if _use_log(self.log):
            from inform import log
            log('running:', render_command(cmd, option_args=self.option_args))

        if self.save_stdout or self.save_stderr:
            try:
                DEVNULL = subprocess.DEVNULL
            except AttributeError:
                DEVNULL = open(os.devnull, 'wb')
        assert self.merge_stderr_into_stdout is False, 'M not supported, use E'

        streams = {}
        if stdin is not None:
            streams['stdin'] = subprocess.PIPE
        if self.save_stdout:
            streams['stdout'] = DEVNULL
        if self.save_stderr:
            streams['stderr'] = DEVNULL

        # run the command
        process = subprocess.Popen(
            cmd, shell=self.use_shell, env=self.env, **streams
        )

        # store needed information and wait for termination if desired
        self.pid = process.pid
        self.process = process

        # write to stdin
        if stdin is not None:
            process.stdin.write(stdin.encode(self.encoding))
            process.stdin.close()

    # wait {{{3
    def wait(self):
        """
        Wait for command to terminate.

        This should only be used if wait-for-termination is False.

        Returns exit status of the command.
        """
        process = self.process

        stdin = self.stdin if self.stdin else ''
        stdout, stderr = process.communicate(stdin.encode(self.encoding))
        self.stdout = None if stdout is None else stdout.decode(self.encoding)
        self.stderr = None if stderr is None else stderr.decode(self.encoding)
        self.status = process.returncode

        if _use_log(self.log):
            from inform import log
            log('exit status:', self.status)

        # check return code
        if self.accept.unacceptable(self.status):
            if self.stderr:
                msg = self.stderr.strip()
            else:
                msg = 'unexpected exit status (%d)' % self.status
            if PREFERENCES['use_inform']:
                from inform import Error
                raise Error(
                    msg = msg,
                    status = self.status,
                    stdout = self.stdout.rstrip() if self.stdout else None,
                    stderr = self.stderr.rstrip() if self.stderr else None,
                    cmd = render_command(self.cmd),
                    template = '{msg}'
                )
            else:
                raise OSError(None, msg)
        return self.status

    # kill {{{3
    def kill(self):
        self.process.kill()
        self.process.wait()

    # __str__ {{{3
    def __str__(self):
        if is_str(self.cmd):
            return self.cmd
        else:
            return ' '.join(str(c) for c in self.cmd)


# Run class {{{2
class Run(Cmd):
    """Run a command immediately.

    See Cmd for information on the arguments.
    Default mode is 'soeW0'.

    Common Examples:
       Run command without capturing stdout and stderr:
           Run(['grep', filename], modes='soeW1')
       Run command and capture stdout; stderr is not captured:
           output = Run(['grep', filename], modes='sOeW1').stdout
       Run command and capture stdout; merge stderr into stdout:
           output = Run(['grep', filename], modes='sOMW1').stdout
    """
    def __init__(
        self, cmd, modes=None, stdin=None, env=None, encoding=None,
        log=None, option_args=None
    ):
        self.cmd = cmd
        self.stdin = None
        self.use_shell = False
        self.save_stdout = False
        self.save_stderr = False
        self.merge_stderr_into_stdout = False
        self.wait_for_termination = True
        self.accept = (0,)
        self.env = env
        self.encoding = PREFERENCES['encoding'] if not encoding else encoding
        self.log = log
        self.option_args = option_args
        self._interpret_modes(modes)
        self.run(stdin)

# Sh class (deprecated) {{{2
class Sh(Cmd):
    """Run a command immediately in the shell.

    See Cmd for information on the arguments, see Run for examples.
    Sh() is the same as Run except S mode is default.
    Default mode is 'SoeW0'.
    """
    def __init__(
        self, cmd, modes=None, stdin=None, env=None, encoding=None,
        log=None, option_args=None
    ):
        self.cmd = cmd
        self.stdin = None
        self.use_shell = True
        self.save_stdout = False
        self.save_stderr = False
        self.merge_stderr_into_stdout = False
        self.wait_for_termination = True
        self.env = env
        self.encoding = PREFERENCES['encoding'] if not encoding else encoding
        self.log = log
        self.option_args = option_args
        self._interpret_modes(modes)
        self.run(stdin)


# Start class {{{2
class Start(Cmd):
    """Run a command immediately, don't wait for it to exit.

    See Cmd for information on the arguments.
    Start() is the similar to Run when using 'w' mode.  The difference is if you
    specify O or E modes, those streams are suppressed rather than being
    captured.
    """
    def __init__(
        self, cmd, modes=None, stdin=None, env=None, encoding=None,
        log=None, option_args=None
    ):
        self.cmd = cmd
        self.stdin = None
        self.use_shell = False
        self.save_stdout = False
        self.save_stderr = False
        self.merge_stderr_into_stdout = False
        self.wait_for_termination = False
        self.accept = (0,)
        self.env = env
        self.encoding = PREFERENCES['encoding'] if not encoding else encoding
        self.log = log
        self.option_args = option_args
        self._interpret_modes(modes)
        self.start(stdin)


# _Accept class {{{2
class _Accept(object):
    # accept exit codes may be specified as:
    # 1. True or the string '*':
    #       all codes are acceptable, or
    # 2. an integer N or the string 'N':
    #       code must be less than or equal to N, or
    # 3. a tuple of integers or a string of the form 'M,N,...':
    #       code must be one of the numbers
    # 4. 0 or empty string:
    #       the only valid return code is 0 (default)
    def __init__(self, accept=0):
        if is_str(accept):
            if accept == '*':
                accept = True
            else:
                try:
                    codes = tuple([int(n) for n in accept.split(',')])
                    accept = codes[0] if len(codes) == 1 else codes
                except ValueError:
                    if accept:
                        raise AssertionError('invalid modes string')
                    else:
                        accept = 0
        self.accept = accept

    def unacceptable(self, status):
        if self.accept is True:
            return False
        elif type(self.accept) is tuple:
            return status not in self.accept
        else:
            return status < 0 or status > self.accept


# run (deprecated) {{{2
def run(cmd, stdin=None, accept=0, shell=False):
    "Run a command without capturing its output."
    import subprocess

    # I have never been able to get Popen to work properly if cmd is not
    # a string when using the shell
    if shell and not is_str(cmd):
        cmd = ' '.join(to_str(c) for c in cmd)
    elif is_str(cmd) and not shell:
        cmd = cmd.split()

    streams = {} if stdin is None else {'stdin': subprocess.PIPE}
    process = subprocess.Popen(cmd, shell=shell, **streams)
    if stdin is not None:
        process.stdin.write(stdin.encode(PREFERENCES['encoding']))
        process.stdin.close()
    status = process.wait()
    if _Accept(accept).unacceptable(status):
        raise OSError(None, "unexpected exit status (%d)." % status)
    return status

# sh (deprecated) {{{2
def sh(cmd, stdin=None, accept=0, shell=True):
    "Execute a command with a shell without capturing its output"
    return run(cmd, stdin, accept, shell=True)


# bg (deprecated) {{{2
def bg(cmd, stdin=None, shell=False):
    "Execute a command in the background without capturing its output."
    import subprocess
    streams = {'stdin': subprocess.PIPE} if stdin is not None else {}
    process = subprocess.Popen(cmd, shell=shell, **streams)
    if stdin is not None:
        process.stdin.write(stdin.encode(PREFERENCES['encoding']))
        process.stdin.close()
    return process.pid

# shbg (deprecated) {{{2
def shbg(cmd, stdin=None, shell=True):
    "Execute a command with a shell in the background without capturing its output."
    return bg(cmd, stdin, shell=True)


# which {{{2
def which(name, path=None, flags=os.X_OK):
    "Search PATH for executable files with the given name."
    result = []
    if path is None:
        path = os.environ.get('PATH', '')
    for p in path.split(os.pathsep):
        p = os.path.join(p, name)
        if os.access(p, flags):
            result.append(p)
    return result

# render_command {{{2
def render_command(cmd, option_args=None, width=70):
    """ Render a command.

    Converts the command to a string.  The formatting is such that you should be
    able to feed the result directly to a shell and have command execute
    properly.

    *cmd* is the command to render. It may be a string or a list of strings.

    *option_args* is a dictionary.  The keys are options accepted by the command
    and the value is the number of arguments for that option.  If an option is
    not found, it is assumed to have 0 arguments.

    *width* specifies how long the string must be before it is broken into
    multiple lines.  If length of resulting line would be width or less, return
    as a single line, otherwise place each argument and option on separate line.

    If the command is rendered as multiple lines, each argument and option is
    placed on a separate line, while keeping argument to options on the same
    line as the option.  Placing each option and argument on its own line allows
    complicated commands with long arguments to be displayed cleanly.

    For example::

    >>> args = {'--dux': 2, '-d': 2, '--tux': 1}
    >>> print(render_command('bux --dux a b -d c d --tux e f g h', args))
    bux --dux a b -d c d --tux e f g h

    >>> print(render_command('bux --dux a b -d c d --tux e f g h', args, width=0))
    bux \
        --dux a b \
        -d c d \
        --tux e \
        f \
        g \
        h

    """
    if option_args is None:
        option_args = {}

    if is_str(cmd):
        components = split_cmd(cmd)
    else:
        components = cmd[:]
        cmd = ' '.join(str(c) for c in components)
    if len(cmd) <= width:
        return cmd

    components.reverse()
    lines = []
    while components:
        opt = components.pop()
        num_args = option_args.get(opt, 0)
        argument = [quote_arg(opt)]
        for i in range(num_args):
            argument.append(quote_arg(components.pop()))
        lines.append(' '.join(argument))
    return ' \\\n    '.join(lines)



# Script

# License {{{1
# Copyright (C) 2016-18 Kenneth S. Kundert
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see http://www.gnu.org/licenses/.


# Imports {{{1
from .dialog import show_list_dialog
from .error import PasswordError
from textwrap import dedent
import re


# Script class {{{1
class Script:
    """Script

    Takes a string that contains attributes. Those attributes are expanded
    before being output. For example::

        Script('username: {username}, password: {passcode}')

    In this case, *{username}* and *{passcode}* are replaced by with the value
    of the corresponding account attribute. In addition to the account
    attributes, *{tab}* and *{return}* are replaced by a tab or carriage return
    character.

    Args:
        script (str):  The script.
    """
    SPLITTER = re.compile(r'({[\w. ]+})')

    def __init__(self, script='username: {username}, password: {passcode}'):
        self.script = script
        self.account = None
        self.rendered = None
        self.is_secret = None

    def initialize(self, account, field_name=None, field_key=None):
        self.account = account

    def __str__(self):
        if self.rendered is not None:
            return self.rendered
        self.is_secret = False
        self.rendered = ''.join(val for cmd, val in self.components())
        return self.rendered

    def components(self, ask=False):
        """Iterates through the script.

        Yields a tuple for each component of a script. The tuple consists of the
        type of the component, and the value of the component.  The type may be 
        'tab' (a tab character), 'return' (a return character), 'text' (raw
        text), 'value' (the value of a field that is not a secret), 'sleep N' (a
        request to sleep N seconds), and finally a field name (the value of a
        field that is secret)..

        Args:
            ask (bool):
                Determines what happens when a composite field is encountered.
                If ask is true a dialog window is opened that that allows the
                user to select the desired member of the collection. If false, a
                PasswordError is raised.

        Raises:
            :exc:`avendesora.PasswordError`: attribute not found.
        """
        account = self.account
        for term in self.SPLITTER.split(self.script):
            if term and term[0] == '{' and term[-1] == '}':
                # we have found a command
                cmd = term[1:-1].lower()

                if cmd == 'tab':
                    val = '\t'
                elif cmd == 'return':
                    val = '\n'
                elif cmd.startswith('sleep '):
                    val = ''
                else:
                    name, key = account.split_field(cmd)
                    try:
                        value = account.get_scalar(name, key)
                    except PasswordError as e:
                        if ask and e.is_collection and len(e.collection):
                            # is composite value, ask user which one is desired
                            choices = e.collection
                            if len(choices) == 1:
                                choice = choices.keys()[0]
                            else:
                                choice = show_list_dialog(
                                    'Choose from %s' % name,
                                    sorted(choices.keys())
                                )
                                if choice is None:
                                    raise PasswordError('user abort.',
                                    culprit=(
                                        account.get_name(),
                                        account.combine_field(name, key)
                                    )
                                )
                            key = choices[choice]
                            value = account.get_scalar(name, key)
                        else:
                            raise
                    val = dedent(str(value)).strip()
                    if not account.is_secret(name, key):
                        cmd = 'value'
            else:
                cmd = 'text'
                val = term
                if not term:
                    continue
            yield cmd, val

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.script)


